from .mathtable import MathTable
from .styles import styledchr, styledstr
from .zmath import Math, Latex, Text, declareoperator
from .config import config

__version__ = '0.8.1'


declareoperator(r'\tg')
declareoperator(r'\ctg')
declareoperator(r'\arcctg')
declareoperator(r'\arctg')
declareoperator(r'\arg')
declareoperator(r'\cotg')
declareoperator(r'\sh')
declareoperator(r'\ch')
declareoperator(r'\cth')
declareoperator(r'\th')



''' <mfrac> math element '''
from xml.etree import ElementTree as ET

from ziafont.fonttypes import BBox

from ..styles import parse_style
from ..drawable import HLine
from . import Mnode


class Mfrac(Mnode, tag='mfrac'):
    ''' Fraction node '''
    # TODO: bevelled attribute for x/y fractions with slanty bar
    def __init__(self, element: ET.Element, parent: 'Mnode', **kwargs):
        pre_style = parse_style(element, parent.style)

        # check original mml attribute for displaystyle to see if
        # it was explicitly turned on (eg dfrac) and not inherited
        if (element.attrib.get('displaystyle') != 'true'
            and ('sup' in kwargs
                 or 'sub' in kwargs
                 or 'frac' in kwargs
                 or not pre_style.displaystyle)):
            element.set('scriptlevel', str(pre_style.scriptlevel + 1))

        # super() after determining scriptlevel so that scale factors are calculated
        super().__init__(element, parent, **kwargs)
        assert len(self.element) == 2
        kwargs['frac'] = True
        kwargs.pop('sup', None)
        self.numerator = Mnode.fromelement(self.element[0], parent=self, **kwargs)
        self.denominator = Mnode.fromelement(self.element[1], parent=self, **kwargs)
        self._setup(**kwargs)

    def _setup(self, **kwargs) -> None:

        linethick = self.units_to_points(self.font.math.consts.fractionRuleThickness)
        if 'linethickness' in self.element.attrib:
            lt = self.element.get('linethickness', '')
            try:
                linethick = self.size_px(lt)
            except ValueError:
                linethick = {'thin': linethick * .5,
                             'thick': linethick * 2}.get(lt, linethick)

        if self.style.displaystyle:
            ynum = self.units_to_points(
                -self.font.math.consts.fractionNumeratorDisplayStyleShiftUp)
            ydenom = self.units_to_points(
                self.font.math.consts.fractionDenominatorDisplayStyleShiftDown)
            numgap = self.units_to_points(
                self.font.math.consts.fractionNumDisplayStyleGapMin)
            denomgap = self.units_to_points(
                self.font.math.consts.fractionDenomDisplayStyleGapMin)
        else:
            ynum = self.units_to_points(
                -self.font.math.consts.fractionNumeratorShiftUp)
            ydenom = self.units_to_points(
                self.font.math.consts.fractionDenominatorShiftDown)
            numgap = self.units_to_points(
                self.font.math.consts.fractionNumeratorGapMin)
            denomgap = self.units_to_points(
                self.font.math.consts.fractionDenominatorGapMin)

        denombox = self.denominator.bbox
        numbox = self.numerator.bbox
        if self.parent.mtag == 'mrow':
            # Make sure axisheight aligns across mrow even with different font sizes
            axheight = self.parent.units_to_points(self.font.math.consts.axisHeight)
        else:
            axheight = self.units_to_points(self.font.math.consts.axisHeight)

        ynum = min(ynum, -(axheight + numgap - numbox.ymin + linethick/2))
        ydenom = max(ydenom, (-axheight + denomgap + denombox.ymax + linethick/2))

        x = 0.
        if (leftsibling := self.leftsibling()):
            if leftsibling.mtag == 'mfrac':
                x = self.size_px('verythinmathspace')
            else:
                x = self.size_px('thinmathspace')

        width = max(numbox.xmax, denombox.xmax)
        xnum = x + (width - (numbox.xmax - numbox.xmin))/2
        xdenom = x + (width - (denombox.xmax - denombox.xmin))/2
        self.nodes.append(self.numerator)
        self.nodes.append(self.denominator)
        self.nodexy.append((xnum, ynum))
        self.nodexy.append((xdenom, ydenom))

        bary = -axheight - linethick/2
        self.nodes.append(HLine(width, linethick, style=self.style, **kwargs))
        self.nodexy.append((x, bary))

        # Calculate/cache bounding box
        xmin = 0
        xmax = x + max(numbox.xmax, denombox.xmax)
        xmax += self.size_px('thinmathspace')
        ymin = (-ydenom) + denombox.ymin
        ymax = (-ynum) + numbox.ymax
        self.bbox = BBox(xmin, xmax, ymin, ymax)



''' Math Nodes '''
from .mnode import Mnode
from .mrow import Mrow, Merror
from .menclose import Menclose
from .mfenced import Mfenced
from .mfrac import Mfrac
from .mnumber import Midentifier, Mnumber, Mtext
from .moperator import Moperator
from .mroot import Mroot, Msqrt
from .mspace import Mpadded, Mphantom, Mspace
from .msubsup import Msub, Msup, Msubsup, Mmultiscripts
from .munderover import Munder, Mover, Munderover
from .mtable import Mtable



''' <msub>, <msup>, <msubsup> Superscript and Subscript Elements '''
from __future__ import annotations
import xml.etree.ElementTree as ET

from ziafont.fonttypes import BBox

from ..mathfont import MathFont
from .. import operators
from .mnode import Mnode


def place_super(base: Mnode, superscript: Mnode, font: MathFont) -> tuple[float, float, float]:
    ''' Superscript. Can be above the operator (like sum) or regular super '''
    if base.params.get('movablelimits') == 'true' and base.style.displaystyle:
        x = (-(base.bbox.xmax - base.bbox.xmin) / 2
             - (superscript.bbox.xmax - superscript.bbox.xmin) / 2)
        supy = (-base.bbox.ymax
                - base.units_to_points(font.math.consts.upperLimitGapMin)
                + superscript.bbox.ymin)
        xadvance = 0.
    else:
        x = 0.
        if base.params.get('movablelimits') == 'true':
            x -= base.size_px(base.params.get('rspace', '0'))

        lastg = base.lastglyph()
        firstg = superscript.firstglyph()

        if lastg:
            if ((italicx := font.math.italicsCorrection.getvalue(lastg.index))
                    and base.lastchar() not in operators.integrals):
                x += base.units_to_points(italicx)

        shiftup = max(font.math.consts.superscriptShiftUp,
                      -firstg.bbox.ymin + font.math.consts.superscriptBottomMin if firstg else 0,
                      base.points_to_units(base.bbox.ymax) - font.math.consts.superscriptBaselineDropMax)

        if superscript.mtag in ['mi', 'mn']:
            if firstg and lastg and lastg.index >= 0 and font.math.kernInfo:  # assembled glyphs have idx<0
                kern, shiftup = font.math.kernsuper(lastg, firstg)
                x += base.units_to_points(kern)

        supy = base.units_to_points(-shiftup)
        xadvance = x + superscript.bbox.xmax
    return x, supy, xadvance


def place_sub(base: Mnode, subscript: Mnode, font: MathFont) -> tuple[float, float, float]:
    ''' Calculate subscript. Can be below the operator (like sum) or regular sub '''
    if base.params.get('movablelimits') == 'true' and base.style.displaystyle:
        x = -(base.bbox.xmax - base.bbox.xmin) / 2 - (subscript.bbox.xmax - subscript.bbox.xmin) / 2
        suby = (-base.bbox.ymin
                + base.units_to_points(font.math.consts.lowerLimitGapMin)
                + subscript.bbox.ymax)
        xadvance = 0.
    else:
        x = 0.

        if base.params.get('movablelimits') == 'true':
            x -= base.size_px(base.params.get('rspace', '0'))

        lastg = base.lastglyph()
        firstg = subscript.firstglyph()

        if lastg:
            if ((italicx := font.math.italicsCorrection.getvalue(lastg.index))
                    and base.lastchar() in operators.integrals):
                x -= base.units_to_points(italicx)  # Shift back on integrals

        if lastg and firstg and lastg.index > 0 and font.math.kernInfo:
            # Horizontal kern set by font
            # Ignoring vertical kern otherwise subscripts don't line up
            # across the row which looks weird
            kern, _ = font.math.kernsub(lastg, firstg)
            x += base.units_to_points(kern)

        if base.mtag in ['mi', 'mn'] or (base.mtag == 'mo' and not base.string):  # type: ignore
            shiftdn = font.math.consts.subscriptShiftDown
        else:
            shiftdn = max(font.math.consts.subscriptShiftDown,
                      firstg.bbox.ymax - font.math.consts.subscriptTopMax if firstg else 0,
                      font.math.consts.subscriptBaselineDropMin - base.points_to_units(base.bbox.ymin))
        suby = base.units_to_points(shiftdn)
        xadvance = x + subscript.xadvance()
    return x, suby, xadvance


class Msup(Mnode, tag='msup'):
    ''' Superscript Node '''
    def __init__(self, element: ET.Element, parent: 'Mnode', **kwargs):
        super().__init__(element, parent, **kwargs)
        assert len(self.element) == 2
        self.base = Mnode.fromelement(self.element[0], parent=self, **kwargs)
        kwargs['sup'] = True
        self.increase_child_scriptlevel(self.element[1])
        self.superscript = Mnode.fromelement(self.element[1], parent=self, **kwargs)
        self._setup(**kwargs)

    def _setup(self, **kwargs) -> None:
        self.nodes.append(self.base)
        self.nodexy.append((0, 0))
        x = self.base.xadvance()

        supx, supy, xadv = place_super(self.base, self.superscript, self.font)
        self.nodes.append(self.superscript)
        self.nodexy.append((x+supx, supy))
        if self.base.bbox.ymax > self.base.bbox.ymin:
            xmin = min(self.base.bbox.xmin, x+supx+self.superscript.bbox.xmin)
            xmax = max(x + xadv, self.base.bbox.xmax, x+supx+self.superscript.bbox.xmax)
            ymin = min(self.base.bbox.ymin, -supy + self.superscript.bbox.ymin)
            ymax = max(self.base.bbox.ymax, -supy + self.superscript.bbox.ymax)
        else:  # Empty base
            xmin = self.superscript.bbox.xmin
            ymin = -supy
            xmax = x + xadv
            ymax = -supy + self.superscript.bbox.ymax
        self.bbox = BBox(xmin, xmax, ymin, ymax)


class Msub(Mnode, tag='msub'):
    ''' Subscript Node '''
    def __init__(self, element: ET.Element, parent: 'Mnode', **kwargs):
        super().__init__(element, parent, **kwargs)
        assert len(self.element) == 2
        self.base = Mnode.fromelement(self.element[0], parent=self, **kwargs)
        kwargs['sub'] = True
        self.increase_child_scriptlevel(self.element[1])
        self.subscript = Mnode.fromelement(self.element[1], parent=self, **kwargs)
        self._setup(**kwargs)

    def _setup(self, **kwargs) -> None:
        self.nodes.append(self.base)
        self.nodexy.append((0, 0))
        x = self.base.xadvance()

        subx, suby, xadv = place_sub(self.base, self.subscript, self.font)
        self.nodes.append(self.subscript)
        self.nodexy.append((x + subx, suby))

        xmin = min(self.base.bbox.xmin, x+subx+self.subscript.bbox.xmin)
        xmax = max(x + xadv, self.base.bbox.xmax, x+subx+self.subscript.bbox.xmax)
        ymin = min(self.base.bbox.ymin, -suby+self.subscript.bbox.ymin)
        ymax = max(self.base.bbox.ymax, -suby+self.subscript.bbox.ymax)
        self.bbox = BBox(xmin, xmax, ymin, ymax)


class Msubsup(Mnode, tag='msubsup'):
    ''' Subscript and Superscript together '''
    def __init__(self, element: ET.Element, parent: 'Mnode', **kwargs):
        super().__init__(element, parent, **kwargs)
        assert len(self.element) == 3
        self.base = Mnode.fromelement(self.element[0], parent=self, **kwargs)
        kwargs['sup'] = True
        kwargs['sub'] = True
        self.increase_child_scriptlevel(self.element[1])
        self.increase_child_scriptlevel(self.element[2])
        self.subscript = Mnode.fromelement(
            self.element[1], parent=self, **kwargs)
        self.superscript = Mnode.fromelement(
            self.element[2], parent=self, **kwargs)
        self._setup(**kwargs)

    def _setup(self, **kwargs) -> None:
        self.nodes.append(self.base)
        self.nodexy.append((0, 0))
        x = self.base.xadvance()
        subx, suby, xadvsub = place_sub(self.base, self.subscript, self.font)
        supx, supy, xadvsup = place_super(self.base, self.superscript, self.font)

        # Ensure subSuperscriptGapMin between scripts
        if ((suby - self.subscript.bbox.ymax) - (supy-self.superscript.bbox.ymin)
                < self.units_to_points(self.font.math.consts.subSuperscriptGapMin)):
            diff = (self.units_to_points(self.font.math.consts.subSuperscriptGapMin)
                    - (suby - self.subscript.bbox.ymax)
                    + (supy-self.superscript.bbox.ymin))
            suby += diff/2
            supy -= diff/2

        self.nodes.append(self.subscript)
        self.nodexy.append((x + subx, suby))
        self.nodes.append(self.superscript)
        self.nodexy.append((x + supx, supy))

        if self.base.bbox.ymax > self.base.bbox.ymin:
            xmin = self.base.bbox.xmin
            xmax = max(x + xadvsup, x + xadvsub)
            ymin = min(-self.base.bbox.ymin, -suby + self.subscript.bbox.ymin)
            ymax = max(self.base.bbox.ymax, -supy + self.superscript.bbox.ymax)
        else:  # Empty base
            xmin = 0
            ymin = -suby
            xmax = x + max(xadvsub, xadvsup)
            ymax = -supy + self.superscript.bbox.ymax

        self.bbox = BBox(xmin, xmax, ymin, ymax)


class Mmultiscripts(Mnode, tag='mmultiscripts'):
    ''' Multiple sub/superscripts in one element '''
    def __init__(self, element: ET.Element, parent: 'Mnode', **kwargs):
        super().__init__(element, parent, **kwargs)

        self.base = Mnode.fromelement(self.element[0], parent=self, **kwargs)
        self.prescripts: list[ET.Element] = []
        self.postscripts: list[ET.Element] = []

        activescriptlist = self.postscripts
        for elm in self.element[1:]:
            if elm.tag == 'mprescripts':
                activescriptlist = self.prescripts
                continue
            activescriptlist.append(elm)
        self._setup(**kwargs)

    def _setup(self, **kwargs):
        ysub = self.units_to_points(self.font.math.consts.subscriptShiftDown)
        ysup = -self.units_to_points(self.font.math.consts.superscriptShiftUp)

        x = xmax = 0
        ymax = self.base.bbox.ymax
        ymin = self.base.bbox.ymin

        for sub, sup in zip(self.prescripts[::2], self.prescripts[1::2]):
            self.increase_child_scriptlevel(sub)
            self.increase_child_scriptlevel(sup)
            subnode = Mnode.fromelement(sub, parent=self, **kwargs)
            supnode = Mnode.fromelement(sup, parent=self, **kwargs)
            width = max(subnode.bbox.xmax, supnode.bbox.xmax)
            self.nodes.append(subnode)
            self.nodexy.append((x+width-subnode.bbox.xmax, ysub))
            self.nodes.append(supnode)
            self.nodexy.append((x+width-supnode.bbox.xmax, ysup))
            
            ymax = max(ymax, -ysup+supnode.bbox.ymax)
            ymin = min(ymin, -ysub+subnode.bbox.ymin)
            xmax = max(xmax, x+width)
            x += width
            x += self.units_to_points(self.font.math.consts.spaceAfterScript)

        self.nodes.append(self.base)
        self.nodexy.append((x, 0))
        x += self.base.xadvance()
        xmax = max(xmax, x+self.base.bbox.xmax)

        for sub, sup in zip(self.postscripts[::2], self.postscripts[1::2]):
            self.increase_child_scriptlevel(sub)
            self.increase_child_scriptlevel(sup)
            subnode = Mnode.fromelement(sub, parent=self, **kwargs)
            supnode = Mnode.fromelement(sup, parent=self, **kwargs)
            self.nodes.append(subnode)
            self.nodexy.append((x, ysub))
            self.nodes.append(supnode)
            self.nodexy.append((x, ysup))

            ymax = max(ymax, -ysup+supnode.bbox.ymax)
            ymin = min(ymin, -ysub+subnode.bbox.ymin)
            xmax = max(xmax, x+subnode.bbox.xmax, x+supnode.bbox.xmax)
            x += max(subnode.xadvance(), supnode.xadvance())
            x += self.units_to_points(self.font.math.consts.spaceAfterScript)

        self.bbox = BBox(0, xmax, ymin, ymax)



''' Apply italic, bold, and other font styles by shifting the unstyled ASCII
    characters [A-Z, a-z, and 0-9] to their higher unicode alternatives. Note
    this does not check whether the new character glyph exists in the font.
'''
from __future__ import annotations
from typing import Any, MutableMapping
from collections import ChainMap, namedtuple
from dataclasses import dataclass, field, asdict
from xml.etree import ElementTree as ET

from .config import config


VARIANTS = ['serif', 'sans', 'script', 'double', 'mono', 'fraktur']
Styletype = namedtuple('Styletype', 'bold italic')


@dataclass
class MathVariant:
    ''' Math font variant, such as serif, sans, script, italic, etc. '''
    style: str = 'serif'
    italic: bool = False
    bold: bool = False
    normal: bool = False


@dataclass
class MathStyle:
    ''' Math Style parameters '''
    mathvariant: MathVariant = field(default_factory=MathVariant)
    displaystyle: bool = True
    mathcolor: str = 'black'
    mathbackground: str = 'none'
    mathsize: str = ''
    scriptlevel: int = 0


def parse_variant(variant: str, parent_variant: MathVariant) -> MathVariant:
    ''' Extract mathvariant from MathML attribute and parent's variant '''
    bold = True if 'bold' in variant else parent_variant.bold
    italic = True if 'italic' in variant else parent_variant.italic
    normal = True if 'normal' in variant else parent_variant.normal

    variant = variant.replace('bold', '').replace('italic', '').strip()
    if variant in VARIANTS:
        style = variant
    else:
        style = parent_variant.style

    return MathVariant(style=style, italic=italic, bold=bold, normal=normal)


def parse_displaystyle(params: MutableMapping[str, Any]) -> bool:
    ''' Extract displaystyle mode from MathML attributes '''
    dstyle = True
    if 'displaystyle' in params:
        dstyle = params.get('displaystyle') in ['true', True]
    elif 'display' in params:
        dstyle = params.get('display', 'block') != 'inline'
    return dstyle


def parse_style(element: ET.Element, parent_style: MathStyle = None) -> MathStyle:
    ''' Read element style attributes into MathStyle '''
    params: MutableMapping[str, Any]
    if parent_style:
        params = ChainMap(element.attrib, asdict(parent_style))
        parent_variant = parent_style.mathvariant
    else:
        params = element.attrib
        parent_variant = MathVariant()

    args: dict[str, Any] = {}
    args['mathcolor'] = params.get('mathcolor', config.math.color)
    args['mathbackground'] = params.get('mathbackground', config.math.background)
    args['mathsize'] = params.get('mathsize', '')
    args['scriptlevel'] = int(params.get('scriptlevel', 0))
    args['mathvariant'] = parse_variant(element.attrib.get('mathvariant', config.math.variant), parent_variant) 
    args['displaystyle'] = parse_displaystyle(params)
    
    css = params.get('style', '')
    if css:
        cssparams = css.split(';')
        for cssparam in cssparams:
            if not cssparam:
                continue  # blank lines
            key, val = cssparam.split(':')
            key = key.strip()
            val = val.strip()
            if key.lower() == 'background':
                args['mathbackground'] = val
            elif key.lower() == 'color':
                args['mathcolor'] = val
    return MathStyle(**args)


LATIN_CAP_RANGE = (0x41, 0x5A)
LATIN_CAPS = \
    {'serif': {Styletype(bold=False, italic=False): 0x0000,
               Styletype(bold=True, italic=False): 0x1D400,
               Styletype(bold=False, italic=True): 0x1D434,
               Styletype(bold=True, italic=True): 0x1D468
               },
     'sans':  {Styletype(bold=False, italic=False): 0x1D5A0,
               Styletype(bold=True, italic=False): 0x1D5D4,
               Styletype(bold=False, italic=True): 0x1D608,
               Styletype(bold=True, italic=True): 0x1D63C
               },
     'script': {Styletype(bold=False, italic=False): 0x1D49C,
                Styletype(bold=True, italic=False): 0x1D4D0,
                Styletype(bold=True, italic=True): 0x1D4D0  # No separate italic
                },
     'fraktur': {Styletype(bold=False, italic=False): 0x1D504,
                 Styletype(bold=True, italic=False): 0x1D56C, 
                 Styletype(bold=True, italic=True): 0x1D56C  # No separate italic
                 },
     'mono': {Styletype(bold=False, italic=False): 0x1D670,
              },
     'double': {Styletype(bold=False, italic=False): 0x1D538,
                },
    }

LATIN_SMALL_RANGE = (0x61, 0x7a)
LATIN_SMALL = \
    {'serif': {Styletype(bold=False, italic=False): 0x0000,
               Styletype(bold=True, italic=False): 0x1D41A,
               Styletype(bold=False, italic=True): 0x1D44E,
               Styletype(bold=True, italic=True): 0x1D482
               },
     'sans':  {Styletype(bold=False, italic=False): 0x1D5BA,
               Styletype(bold=True, italic=False): 0x1D5EE,
               Styletype(bold=False, italic=True): 0x1D622,
               Styletype(bold=True, italic=True): 0x1D656
               },
     'script': {Styletype(bold=False, italic=False): 0x1D4B6,
                Styletype(bold=True, italic=False): 0x1D4EA,
                Styletype(bold=True, italic=True): 0x1D4EA  # No separate italic
                },
     'fraktur': {Styletype(bold=False, italic=False): 0x1D51E,
                 Styletype(bold=True, italic=False): 0x1D586,
                 Styletype(bold=True, italic=True): 0x1D586  # No separate italic
                 },
     'mono': {Styletype(bold=False, italic=False): 0x1D68A,
              },
     'double': {Styletype(bold=False, italic=False): 0x1D552,
                },
    }

GREEK_CAP_RANGE = (0x0391, 0x3AA)
GREEK_CAPS = \
    {'serif': {Styletype(bold=False, italic=False): 0x0000,
               Styletype(bold=True, italic=False): 0x1D6A8,
               Styletype(bold=False, italic=True): 0x1D6E2,
               Styletype(bold=True, italic=True): 0x1D71C
               },
     'sans': {Styletype(bold=False, italic=False): 0x0000,
              Styletype(bold=True, italic=False): 0x1D756,
              Styletype(bold=True, italic=True): 0x1D790
              },
    }

GREEK_LOWER_RANGE = (0x3b1, 0x3d0)
GREEK_LOWER = \
    {'serif': {Styletype(bold=False, italic=False): 0x0000,
               Styletype(bold=True, italic=False): 0x1D6C2,
               Styletype(bold=False, italic=True): 0x1D6FC,
               Styletype(bold=True, italic=True): 0x1D736
               },
     'sans': {Styletype(bold=False, italic=False): 0x0000,
              Styletype(bold=True, italic=False): 0x1D770,
              Styletype(bold=True, italic=True): 0x1D7AA
              },
    }

DIGIT_RANGE = (0x30, 0x39)
DIGITS = \
    {'serif': {Styletype(bold=False, italic=False): 0x0000,
               Styletype(bold=True, italic=False): 0x1D7CE,
               },
     'double': {Styletype(bold=False, italic=False): 0x1D7D8, },
     'mono': {Styletype(bold=False, italic=False): 0x1D7F6, },
     'sans': {Styletype(bold=False, italic=False): 0x1D7E2,
              Styletype(bold=True, italic=False): 0x1D7EC,
              Styletype(bold=True, italic=True): 0x1D7EC,
              },
    }


subtables = ((LATIN_CAP_RANGE, LATIN_CAPS),
             (LATIN_SMALL_RANGE, LATIN_SMALL),
             (GREEK_CAP_RANGE, GREEK_CAPS),
             (GREEK_LOWER_RANGE, GREEK_LOWER),
             (DIGIT_RANGE, DIGITS))


# These are the yellow characters in wikipedia's table
OFFSET_EXCEPTIONS = {
    'ϴ': 0x0391+0x11,
    '∇': 0x0391+0x19,
    '∂': 0x03B1+0x19,
    'ϵ': 0x03B1+0x1A,
    'ϑ': 0x03B1+0x1B,
    'ϰ': 0x03B1+0x1C,
    'ϕ': 0x03B1+0x1D,
    'ϱ': 0x03B1+0x1E,
    'ϖ': 0x03B1+0x1F}


EXCEPTIONS = {
    0x1D49C+0x01: 'ℬ',  # latin cap scripts
    0x1D49C+0x04: 'ℰ',
    0x1D49C+0x05: 'ℱ',
    0x1D49C+0x07: 'ℋ',
    0x1D49C+0x08: 'ℐ',
    0x1D49C+0x0B: 'ℒ',
    0x1D49C+0x0C: 'ℳ',
    0x1D49C+0x11: 'ℛ',
    0x1D504+0x02: 'ℭ',  # latin cap frakturs
    0x1D504+0x07: 'ℌ',
    0x1D504+0x08: 'ℑ',
    0x1D504+0x11: 'ℜ',
    0x1D504+0x19: 'ℨ',
    0x1D538+0x02: 'ℂ',  # latin cap doubles
    0x1D538+0x07: 'ℍ',
    0x1D538+0x0D: 'ℕ',
    0x1D538+0x0F: 'ℙ',
    0x1D538+0x10: 'ℚ',
    0x1D538+0x11: 'ℝ',
    0x1D538+0x19: 'ℤ',
    0x1D44E+0x07: 'ℎ',  # latin small italic
    0x1D4B6+0x04: 'ℯ',  # latin small script
    0x1D4B6+0x06: 'ℊ',
    0x1D4B6+0x0E: 'ℴ',
    }


def auto_italic(char: str) -> bool:
    ''' Determine whether the character should be automatically
        converted to italic
    '''
    ordchr = ord(char)
    for ordrange in (GREEK_LOWER_RANGE, LATIN_SMALL_RANGE, LATIN_CAP_RANGE):
        if ordrange[0] <= ordchr <= ordrange[1]:
            return True
    return False


def styledchr(char, variant: MathVariant):
    ''' Convert character to its styled (bold, italic, script, etc.) variant.
        See tables at: https://en.wikipedia.org/wiki/Mathematical_Alphanumeric_Symbols
    '''
    script = variant.style
    style = Styletype(variant.bold, variant.italic)
    styledchr = char  # Default is to return char unchanged

    charord = OFFSET_EXCEPTIONS.get(char, ord(char))
    for ordrange, table in subtables:
        if ordrange[0] <= charord <= ordrange[1]:
            ordoffset = charord - ordrange[0]
            scripttable = table.get(script, table.get('serif'))
            offset = scripttable.get(style, scripttable.get(Styletype(False, False)))  # type: ignore
            if offset:
                styledchr = chr(ordoffset + offset)
            styledchr = EXCEPTIONS.get(ord(styledchr), styledchr)
            break

    return styledchr


def styledstr(st: str, variant: MathVariant) -> str:
    ''' Apply unicode styling conversion to a string '''
    return ''.join([styledchr(s, variant) for s in st])



''' <mover> and <munder> Math Elements '''
from __future__ import annotations
from typing import Optional, Union
import xml.etree.ElementTree as ET
from copy import copy

from ziafont.fonttypes import BBox
from ziafont.glyph import SimpleGlyph

from ..mathfont import MathFont
from .. import drawable
from .nodetools import elementtext, node_is_singlechar
from . import Mnode


# Accents are drawn same scriptlevel as base
ACCENTS = [
    0x005E,  # \hat, \widehat
    0x02D9,  # \dot
    0x02C7,  # \check
    0x007E,  # \tilde, \widetilde
    0x00B4,  # \acute
    0x0060,  # \grave
    0x00A8,  # \ddot
    0x20DB,  # \dddot
    0x20DC,  # \ddddot
    0x02D8,  # \breve
    0x00AF,  # \bar
    0x02DA,  # \mathring
    ]


def place_over(base: Mnode,
               over: Union[Mnode, drawable.HLine],
               font: MathFont) -> tuple[float, float]:
    ''' Place node over another one

        Args:
            base: Base node
            over: Node to draw over the base
            font: Font

        Returns:
            x, y: position for over node
    '''
    # Center the node by default
    x = (((base.bbox.xmax - base.bbox.xmin) - (over.bbox.xmax-over.bbox.xmin)) / 2
         - over.bbox.xmin)
    
    if ((lastg := base.lastglyph())
            and node_is_singlechar(base)
            and not isinstance(over, drawable.HLine)):
        # Italic adjustment and font-specific accent attachment,
        # if base is a single glyph
        if (italicx := font.math.italicsCorrection.getvalue(lastg.index)):
            x += base.units_to_points(italicx)

        # Use font-specific accent attachment if defined
        if (basex := font.math.topattachment(lastg.index)):
            x = base.units_to_points(basex)
            
            if (node_is_singlechar(over)
                    and (attachx := font.math.topattachment(over.lastglyph().index))):  # type: ignore
                x -= over.units_to_points(attachx)
            else:
                x -= (over.bbox.xmax-over.bbox.xmin)/2

    y = -base.bbox.ymax - base.units_to_points(font.math.consts.overbarVerticalGap)
    y += over.bbox.ymin
    return x, y


def place_under(base: Mnode,
                under: Union[Mnode, drawable.HLine],
                font: MathFont) -> tuple[float, float]:
    ''' Place node under another one
        Args:
            base: Base node
            under: Node to draw under the base
            font: Font

        Returns:
            x, y: position for under node
    '''
    x = (((base.bbox.xmax - base.bbox.xmin) - (under.bbox.xmax-under.bbox.xmin)) / 2
         - under.bbox.xmin)

    if ((lastg := base.lastglyph())
            and node_is_singlechar(base)
            and not isinstance(under, drawable.HLine)):
        if (italicx := font.math.italicsCorrection.getvalue(lastg.index)):
            x -= base.units_to_points(italicx)

    y = -base.bbox.ymin + base.units_to_points(font.math.consts.underbarVerticalGap)
    y += (under.bbox.ymax)
    return x, y


class Mover(Mnode, tag='mover'):
    ''' Over node '''
    # Over/underbar character. Fonts not consistent with using stretchy/assembled
    # glyphs, so draw with HLine instead of glyph.
    BAR = '―'  # 0x2015

    def __init__(self, element: ET.Element, parent: 'Mnode', **kwargs):
        super().__init__(element, parent, **kwargs)
        self.over: Union[drawable.HLine, Mnode]

        kwargs = copy(kwargs)
        assert len(self.element) == 2
        self.base = Mnode.fromelement(self.element[0], parent=self, **kwargs)
        if (self.element[1].tag in ['mover', 'munder']
                or (self.element[1].tag == 'mo' and len(elementtext(self.element[1])) == 1)):
            kwargs['width'] = self.base.bbox.xmax - self.base.bbox.xmin

        self._isaccent = False
        if not (len(elementtext(self.element[1])) == 1
                and ord(elementtext(self.element[1])) in ACCENTS):
            if self.element.get('accent', 'false').lower() == 'false':
                self.increase_child_scriptlevel(self.element[1])
                self._isaccent = True

        if elementtext(self.element[1]) == self.BAR:
            self.over = drawable.HLine(
                kwargs['width'],
                self.units_to_points(self.font.math.consts.overbarRuleThickness))
        else:
            if self.element[1].get('stretchy') == 'true':
                self.element[1].set('lspace', '0')
                self.element[1].set('rspace', '0')
            self.over = Mnode.fromelement(self.element[1], parent=self, **kwargs)

        self._setup(**kwargs)

    def _setup(self, **kwargs) -> None:
        overx, overy = place_over(self.base, self.over, self.font)
        basex = 0.
        if overx < 0:
            basex = -overx
            overx = 0.

        self.nodes.append(self.base)
        self.nodexy.append((basex, 0))
        self.nodes.append(self.over)
        self.nodexy.append((overx, overy))

        xmin = min(overx+self.over.bbox.xmin, basex+self.base.bbox.xmin)
        xmax = max(overx+self.over.bbox.xmax, basex+self.base.bbox.xmax)
        ymin = self.base.bbox.ymin
        ymax = -overy + self.over.bbox.ymax
        self.bbox = BBox(xmin, xmax, ymin, ymax)
        if not self._isaccent:
            self._xadvance = self.base.xadvance()
        else:
            self._xadvance = xmax

    def xadvance(self) -> float:
        return self._xadvance

    def lastglyph(self) -> Optional[SimpleGlyph]:
        ''' Get the last glyph in this node '''
        return self.base.lastglyph()


class Munder(Mnode, tag='munder'):
    ''' Under node '''
    def __init__(self, element: ET.Element, parent: 'Mnode', **kwargs):
        super().__init__(element, parent, **kwargs)
        self.under: Union[drawable.HLine, Mnode]

        kwargs = copy(kwargs)
        assert len(self.element) == 2
        self.base = Mnode.fromelement(self.element[0], parent=self, **kwargs)
        kwargs['sub'] = True
        if (self.element[1].tag in ['mover', 'munder', 'munderover']
                or (self.element[1].tag == 'mo' and len(elementtext(self.element[1])) == 1)):
            kwargs['width'] = self.base.bbox.xmax - self.base.bbox.xmin

        if elementtext(self.element[1]) == Mover.BAR:
            self.under = drawable.HLine(
                kwargs['width'],
                self.units_to_points(self.font.math.consts.underbarRuleThickness))
        else:
            if self.element[1].get('stretchy') == 'true':
                self.element[1].set('lspace', '0')
                self.element[1].set('rspace', '0')
            if self.element.get('accentunder', 'false').lower() == 'false':
                self.increase_child_scriptlevel(self.element[1])
                self._isaccent = True
            self.under = Mnode.fromelement(self.element[1], parent=self, **kwargs)
        self._setup(**kwargs)

    def _setup(self, **kwargs) -> None:
        underx, undery = place_under(self.base, self.under, self.font)

        basex = 0.
        if underx < 0:
            basex = -underx
            underx = 0

        self.nodes.append(self.base)
        self.nodexy.append((basex, 0))
        self.nodes.append(self.under)
        self.nodexy.append((underx, undery))

        xmin = min(underx+self.under.bbox.xmin, basex+self.base.bbox.xmin)
        xmax = max(underx+self.under.bbox.xmax, basex+self.base.bbox.xmax)
        ymin = -undery + self.under.bbox.ymin
        ymax = self.base.bbox.ymax
        self.bbox = BBox(xmin, xmax, ymin, ymax)

    def lastglyph(self) -> Optional[SimpleGlyph]:
        ''' Get the last glyph in this node '''
        return self.base.lastglyph()


class Munderover(Mnode, tag='munderover'):
    ''' Under bar and over bar '''
    def __init__(self, element: ET.Element, parent: 'Mnode', **kwargs):
        super().__init__(element, parent, **kwargs)
        kwargs = copy(kwargs)
        assert len(self.element) == 3
        self.base = Mnode.fromelement(self.element[0], parent=self, **kwargs)
        kwargs['sub'] = True
        if (self.element[1].tag in ['mover', 'munder', 'munderover']
                or (self.element[1].tag == 'mo' and len(elementtext(self.element[1])) == 1)):
            kwargs['width'] = self.base.bbox.xmax - self.base.bbox.xmin
        if self.element.get('accentunder', 'false').lower() == 'false':
            self.increase_child_scriptlevel(self.element[1])

        self.under = Mnode.fromelement(self.element[1], parent=self, **kwargs)

        kwargs.pop('width', None)
        kwargs.pop('sub', None)
        if (self.element[2].tag in ['mover', 'munder']
                or (self.element[2].tag == 'mo' and len(elementtext(self.element[2])) == 1)):
            kwargs['width'] = self.base.bbox.xmax - self.base.bbox.xmin

        if not (len(elementtext(self.element[2])) == 1
                and ord(elementtext(self.element[2])) in ACCENTS):
            if self.element[2].get('stretchy') == 'true':
                self.element[2].set('lspace', '0')
                self.element[2].set('rspace', '0')
            if self.element.get('accent', 'false').lower() == 'false':
                self.increase_child_scriptlevel(self.element[2])

        self.over = Mnode.fromelement(self.element[2], parent=self, **kwargs)
        self._setup(**kwargs)

    def _setup(self, **kwargs) -> None:
        self.nodes = []
        overx, overy = place_over(self.base, self.over, self.font)
        underx, undery = place_under(self.base, self.under, self.font)

        basex = 0.
        if overx < 0 or underx < 0:
            basex = max(-overx, -underx)
            overx, underx = overx - min(overx, underx), underx - min(overx, underx)

        self.nodes.append(self.base)
        self.nodexy.append((basex, 0))
        self.nodes.append(self.over)
        self.nodes.append(self.under)
        self.nodexy.append((overx, overy))
        self.nodexy.append((underx, undery))
        xmin = min(underx, overx, basex+self.base.bbox.xmin)
        xmax = max(underx+self.under.bbox.xmax,
                   overx+self.over.bbox.xmax,
                   basex+self.base.bbox.xmax)
        ymin = -undery + self.under.bbox.ymin
        ymax = -overy + self.over.bbox.ymax
        self.bbox = BBox(xmin, xmax, ymin, ymax)

    def lastglyph(self) -> Optional[SimpleGlyph]:
        ''' Get the last glyph in this node '''
        return self.base.lastglyph()



''' MathML Escape Codes

Python's XML parser doesn't like these, so they must be replaced
before parsing.
'''

import re
from . import escape_codes

ESCAPES = escape_codes.ESCAPES
ESCAPES.update({
    ':=': '≔',
    r'\*=': '⩮',  # NOTE: * is escaped for use in re
    '==': '⩵',
    '!=': '≠',
    '&InvisibleComma;': '',
    '&InvisibleTimes;': '',
    # These are picked up by XML parser already
    # '&lt;' : '<',
    # '&gt;' : '>',
    # '&amp;': '&',
    })

regex = re.compile('|'.join(map(re.escape, ESCAPES.keys())))


def unescape(xmlstr: str) -> str:
    ''' Remove MathML escape codes from xml string '''
    xml = regex.sub(lambda match: ESCAPES[match.group(0)], xmlstr)

    # Replace hyphens with real minus signs, but only within numbers/operators
    # (due to re.escape, the compiled regex used above won't
    #  work for these substitutions)
    xml = re.sub(r'<mn.*>\s*-', '<mn>−', xml)
    xml = re.sub(r'<mo.*>\s*-\s*</mo>', '<mo> − </mo>', xml)
    return xml



''' Main math rendering class '''

from __future__ import annotations
from typing import Union, Literal, Tuple, Optional, Dict
import warnings
import re
from collections import ChainMap
from math import inf, cos, sin, radians
from itertools import zip_longest
import importlib.resources as pkg_resources
import xml.etree.ElementTree as ET

from latex2mathml.converter import convert  # type: ignore
import latex2mathml.commands  # type: ignore

import ziafont as zf
from ziafont.glyph import fmt
from .mathfont import MathFont
from .nodes import Mnode
from .styles import parse_style
from .escapes import unescape
from .config import config


Halign = Literal['left', 'center', 'right']
Valign = Literal['top', 'center', 'base', 'axis', 'bottom']


def denamespace(element: ET.Element) -> ET.Element:
    ''' Recursively remove namespace {...} from beginning of xml
        element names, so they can be searched easily.
    '''
    if element.tag.startswith('{'):
        element.tag = element.tag.split('}')[1]
    for elm in element:
        denamespace(elm)
    return element


def declareoperator(name: str) -> None:
    r''' Declare a new operator name, similar to Latex ``\DeclareMathOperator`` command.

        Args:
            name: Name of operator, should start with a ``\``.
                Example: ``declareoperator(r'\myfunc')``
    '''
    latex2mathml.commands.FUNCTIONS = latex2mathml.commands.FUNCTIONS + (name,)


def tex2mml(tex: str, inline: bool = False) -> str:
    ''' Convert Latex to MathML. Do some hacky preprocessing to work around
        some issues with generated MathML that ziamath doesn't support yet.
    '''
    tex = re.sub(r'\\binom{(.+?)}{(.+?)}', r'\\left( \1 \\atop \2 \\right)', tex)
    # latex2mathml bug requires space after mathrm
    tex = re.sub(r'\\mathrm{(.+?)}', r'\\mathrm {\1}', tex)
    tex = tex.replace('||', '‖')
    if config.decimal_separator == ',':
        # Replace , with {,} to remove right space
        # (must be surrounded by digits)
        tex = re.sub(r'([0-9]),([0-9])', r'\1{,}\2', tex)

    mml = convert(tex, display='inline' if inline else 'block')

    # Replace some operators with "stretchy" variants
    mml = re.sub(r'<mo>&#x0005E;', r'<mo>&#x00302;', mml)  # widehat
    mml = re.sub(r'<mo>&#x0007E;', r'<mo>&#x00303;', mml)  # widetilde
    return mml


def apply_mstyle(element: ET.Element) -> ET.Element:
    ''' Take attributes defined in <mstyle> elements and add them
        to all the child elements, removing the original <mstyle>
    '''
    def flatten_attrib(element: ET.Element) -> None:
        for child in element:
            if element.tag == 'mstyle':
                child.attrib = dict(ChainMap(child.attrib, element.attrib))
            flatten_attrib(child)

    flatten_attrib(element)

    elmstr = ET.tostring(element).decode('utf-8')
    elmstr = re.sub(r'<mstyle.+?>', '', elmstr)
    elmstr = re.sub(r'</mstyle>', '', elmstr)
    return ET.fromstring(elmstr)


class Math:
    ''' MathML Element Renderer

        Args:
            mathml: MathML expression, in string or XML Element
            size: Base font size, pixels
            font: Filename of font file. Must contain MATH typesetting table.
    '''
    def __init__(self, mathml: Union[str, ET.Element],
                 size: float = None, font: str = None):
        self.size = size if size else config.math.fontsize
        font = font if font else config.math.mathfont

        self.font: MathFont
        if font is None:
            self.font = loadedfonts['default']
        elif font in loadedfonts:
            self.font = loadedfonts[font]
        else:
            self.font = MathFont(font, self.size)
            loadedfonts[font] = self.font

        if isinstance(mathml, str):
            mathml = unescape(mathml)
            mathml = ET.fromstring(mathml)
        mathml = denamespace(mathml)
        mathml = apply_mstyle(mathml)

        self.mathml = mathml
        self.style = parse_style(mathml)
        self.element = mathml
        self.mtag = 'math'
        self.node = Mnode.fromelement(mathml, parent=self)  # type: ignore

    @classmethod
    def fromlatex(cls, latex: str, size: float = None, mathstyle: str = None,
                  font: str = None, color: str = None, inline: bool = False):
        ''' Create Math Renderer from a single LaTeX expression. Requires
            latex2mathml Python package.

            Args:
                latex: Latex string
                size: Base font size
                mathstyle: Style parameter for math, equivalent to "mathvariant" MathML attribute
                font: Font file name
                color: Color parameter, equivalent to "mathcolor" attribute
                inline: Use inline math mode (default is block mode)
        '''
        mathml: Union[str, ET.Element]
        mathml = tex2mml(latex, inline=inline)
        if mathstyle:
            mathml = ET.fromstring(mathml)
            mathml.attrib['mathvariant'] = mathstyle
            mathml = ET.tostring(mathml, encoding='unicode')
        if color:
            mathml = ET.fromstring(mathml)
            mathml.attrib['mathcolor'] = color
        return cls(mathml, size, font)

    @classmethod
    def fromlatextext(cls, latex: str, size: float = 24, mathstyle: str = None,
                      textstyle: str = None, font: str = None,
                      color: str = None):
        ''' Create Math Renderer from a sentence containing zero or more LaTeX
            expressions delimited by $..$, resulting in single MathML element.
            Requires latex2mathml Python package.

            Args:
                latex: string
                size: Base font size
                mathstyle: Style parameter for math, equivalent to "mathvariant" MathML attribute
                textstyle: Style parameter for text, equivalent to "mathvariant" MathML attribute
                font: Font file name
                color: Color parameter, equivalent to "mathcolor" attribute
        '''
        warnings.warn(r'fromlatextext is deprecated. Use ziamath.Text or \text{} command.', DeprecationWarning, stacklevel=2)
        
        # Extract each $..$, convert to MathML, but the raw text in <mtext>, and join
        # into a single <math>
        parts = re.split(r'(\$+.*?\$+)', latex)
        texts = parts[::2]
        maths = [tex2mml(p.replace('$', ''), inline=not p.startswith('$$')) for p in parts[1::2]]
        mathels = [ET.fromstring(m) for m in maths]   # Convert to xml, but drop opening <math>

        mml = ET.Element('math')
        for text, mathel in zip_longest(texts, mathels):
            if text:
                mtext = ET.SubElement(mml, 'mtext')
                if textstyle:
                    mtext.attrib['mathvariant'] = textstyle
                mtext.text = text
            if mathel is not None:
                child = mathel[0]
                if mathstyle:
                    child.attrib['mathvariant'] = mathstyle
                if (dstyle := mathel.attrib.get('display')):
                    child.attrib['displaystyle'] = {'inline': 'false',
                                                    'block': 'true'}.get(dstyle, 'true')
                mml.append(child)
        if color:
            mml.attrib['mathcolor'] = color
        return cls(mml, size, font)

    def svgxml(self) -> ET.Element:
        ''' Get standalone SVG of expression as XML Element Tree '''
        svg = ET.Element('svg')
        self.node.draw(1, 0, svg)
        bbox = self.node.bbox
        width = bbox.xmax - bbox.xmin + 2  # Add a 1-px border
        height = bbox.ymax - bbox.ymin + 2

        # Note: viewbox goes negative.
        svg.attrib['width'] = fmt(width)
        svg.attrib['height'] = fmt(height)
        svg.attrib['xmlns'] = 'http://www.w3.org/2000/svg'
        if not config.svg2:
            svg.attrib['xmlns:xlink'] = 'http://www.w3.org/1999/xlink'
        svg.attrib['viewBox'] = f'{fmt(bbox.xmin-1)} {fmt(-bbox.ymax-1)} {fmt(width)} {fmt(height)}'
        return svg

    def drawon(self, svg: ET.Element, x: float = 0, y: float = 0,
               halign: Halign = 'left', valign: Valign = 'base') -> ET.Element:
        ''' Draw the math expression on an existing SVG

            Args:
                x: Horizontal position in SVG coordinates
                y: Vertical position in SVG coordinates
                svg: The image (top-level svg XML object) to draw on
                halign: Horizontal alignment
                valign: Vertical alignment

            Note: Horizontal alignment can be the typical 'left', 'center', or 'right'.
            Vertical alignment can be 'top', 'bottom', or 'center' to align with the
            expression's bounding box, or 'base' to align with the bottom
            of the first text element, or 'axis', aligning with the height of a minus
            sign above the baseline.
        '''
        width, height = self.getsize()
        yshift = {'top': self.node.bbox.ymax,
                  'center': height/2 + self.node.bbox.ymin,
                  'axis': self.node.units_to_points(self.font.math.consts.axisHeight),
                  'bottom': self.node.bbox.ymin}.get(valign, 0)
        xshift = {'center': -width/2,
                  'right': -width}.get(halign, 0)

        svgelm = ET.SubElement(svg, 'g')  # Put it in a group
        self.node.draw(x+xshift, y+yshift, svgelm)
        return svgelm

    def svg(self) -> str:
        ''' Get expression as SVG string '''
        return ET.tostring(self.svgxml(), encoding='unicode')

    def save(self, fname):
        ''' Save expression to SVG file '''
        with open(fname, 'w') as f:
            f.write(self.svg())

    def _repr_svg_(self):
        ''' Jupyter SVG representation '''
        return self.svg()

    @classmethod
    def mathml2svg(cls, mathml: Union[str, ET.Element],
                   size: float = None, font: str = None):
        ''' Shortcut to just return SVG string directly '''
        return cls(mathml, size=size, font=font).svg()

    def getsize(self) -> tuple[float, float]:
        ''' Get size of rendered text '''
        return (self.node.bbox.xmax - self.node.bbox.xmin,
                self.node.bbox.ymax - self.node.bbox.ymin)

    def getyofst(self) -> float:
        ''' Y-shift from bottom of bbox to 0 '''
        return self.node.bbox.ymin


class Latex(Math):
    ''' Render Math from LaTeX

        Args:
            latex: Latex string
            size: Base font size
            mathstyle: Style parameter for math, equivalent to "mathvariant" MathML attribute
            font: Font file name
            color: Color parameter, equivalent to "mathcolor" attribute
            inline: Use inline math mode (default is block mode)
        '''
    def __init__(self, latex: str, size: float = None, mathstyle: str = None,
                 font: str = None, color: str = None, inline: bool = False):
        self.latex = latex

        mathml: Union[str, ET.Element]
        mathml = tex2mml(latex, inline=inline)
        if mathstyle:
            mathml = ET.fromstring(mathml)
            mathml.attrib['mathvariant'] = mathstyle
            mathml = ET.tostring(mathml, encoding='unicode')
        if color:
            mathml = ET.fromstring(mathml)
            mathml.attrib['mathcolor'] = color
        super().__init__(mathml, size, font)


class Text:
    ''' Mixed text and latex math. Inline math delimited by single $..$, and
        display-mode math delimited by double $$...$$. Can contain multiple
        lines. Drawn to SVG.

        Args:
            s: string to write
            textfont: font filename or family name for text
            mathfont: font filename for math
            mathstyle: Style parameter for math
            size: font size in points
            linespacing: spacing between lines
            color: color of text
            halign: horizontal alignment
            valign: vertical alignment
            rotation: Rotation angle in degrees
            rotation_mode: Either 'default' or 'anchor', to
                mimic Matplotlib behavoir. See:
                https://matplotlib.org/stable/gallery/text_labels_and_annotations/demo_text_rotation_mode.html

    '''
    def __init__(self, s, textfont: str = None, mathfont: str = None,
                 mathstyle: str = None, size: float = None, linespacing: float = None,
                 color: str = None,
                 halign: str = 'left', valign: str = 'base',
                 rotation: float = 0, rotation_mode: str = 'anchor'):
        self.str = s
        self.mathfont = mathfont
        self.mathstyle = mathstyle
        self.size = size if size else config.text.fontsize
        self.linespacing = linespacing if linespacing else config.text.linespacing
        self.color = color
        self.textcolor = color if color else config.text.color
        self._halign = halign
        self._valign = valign
        self.rotation = rotation
        self.rotation_mode = rotation_mode
        self.textfont: Optional[Union[MathFont, zf.Font]]

        # textfont can be a path to font, or style type like "serif".
        # If style type, use Stix font variation
        textfont = textfont if textfont else config.text.textfont
        textstyle = config.text.variant
        if textfont is None:
            self.textfont = None
            self.textstyle = textstyle
        else:
            try:
                self.textfont = zf.Font(textfont)
                self.textstyle = textstyle
            except FileNotFoundError:
                self.textfont = None
                self.textstyle = textfont

    def svg(self) -> str:
        ''' Get expression as SVG string '''
        return ET.tostring(self.svgxml(), encoding='unicode')

    def _repr_svg_(self):
        ''' Jupyter SVG representation '''
        return self.svg()

    def svgxml(self) -> ET.Element:
        ''' Get standalone SVG of expression as XML Element Tree '''
        svg = ET.Element('svg')
        _, (x1, x2, y1, y2) = self._drawon(svg)
        svg.attrib['width'] = fmt(x2-x1)
        svg.attrib['height'] = fmt(y2-y1)
        svg.attrib['xmlns'] = 'http://www.w3.org/2000/svg'
        if not config.svg2:
            svg.attrib['xmlns:xlink'] = 'http://www.w3.org/1999/xlink'
        svg.attrib['viewBox'] = f'{fmt(x1)} {fmt(y1)} {fmt(x2-x1)} {fmt(y2-y1)}'
        return svg

    def save(self, fname):
        ''' Save expression to SVG file '''
        with open(fname, 'w') as f:
            f.write(self.svg())

    def drawon(self, svg: ET.Element, x: float = 0, y: float = 0,
               halign: str = None, valign: str = None) -> ET.Element:
        ''' Draw text on existing SVG element

            Args:
                svg: Element to draw on
                x: x-position
                y: y-position
                halign: Horizontal alignment
                valign: Vertical alignment
        '''
        svgelm, _ = self._drawon(svg, x, y, halign, valign)
        return svgelm

    def _drawon(self, svg: ET.Element, x: float = 0, y: float = 0,
                halign: str = None, valign: str = None) -> Tuple[ET.Element, Tuple[float, float, float, float]]:
        ''' Draw text on existing SVG element

            Args:
                svg: Element to draw on
                x: x-position
                y: y-position
                halign: Horizontal alignment
                valign: Vertical alignment
        '''
        halign = self._halign if halign is None else halign
        valign = self._valign if valign is None else valign

        lines = self.str.splitlines()
        svglines = []
        svgelm = ET.SubElement(svg, 'g')

        # Split into lines and "parts"
        linesizes = []
        for line in lines:
            svgparts = []
            parts = re.split(r'(\$+.*?\$+)', line)
            partsizes = []
            for part in parts:
                if not part:
                    continue
                if part.startswith('$$') and part.endswith('$$'):  # Display-mode math
                    math = Math.fromlatex(part.replace('$', ''),
                                          font=self.mathfont,
                                          mathstyle=self.mathstyle,
                                          inline=False,
                                          size=self.size, color=self.color)
                    svgparts.append(math)
                    partsizes.append(math.getsize())

                elif part.startswith('$') and part.endswith('$'):  # Text-mode Math
                    math = Math.fromlatex(part.replace('$', ''),
                                          font=self.mathfont,
                                          mathstyle=self.mathstyle,
                                          inline=True,
                                          size=self.size, color=self.color)
                    svgparts.append(math)
                    partsizes.append(math.getsize())
                else:  # Text
                    if self.textfont:
                        # A specific font file is defined, use ziafont and ignore textstyle
                        txt = zf.Text(part, font=self.textfont, size=self.size, color=self.textcolor)
                        partsizes.append(txt.getsize())
                    else:
                        # use math font with textstyle
                        txt = Math.fromlatex(fr'\text{{{part}}}',
                                             font=self.mathfont,
                                             mathstyle=self.textstyle,
                                             size=self.size,
                                             color=self.textcolor)
                        partsizes.append((txt.node.bbox.xmax - txt.node.bbox.xmin,  # type: ignore
                                          txt.node.size))  # type: ignore
                    svgparts.append(txt)
            if len(svgparts) > 0:
                svglines.append(svgparts)
                linesizes.append(partsizes)

        lineofsts = [min([p.getyofst() for p in line]) for line in svglines]
        lineheights = [max(p[1] for p in line) for line in linesizes]
        linewidths = [sum(p[0] for p in line) for line in linesizes]

        if valign == 'bottom':
            ystart = y + sum(lineofsts) - sum(lineheights[1:])
        elif valign == 'top':
            ystart = y + lineheights[0] + lineofsts[0]
        elif valign == 'center':
            ystart = y + lineheights[0] + lineofsts[0] - sum(lineheights)/2
        else:  # 'base'
            ystart = y

        xmin = ymin = inf
        xmax = ymax = -inf
        yloc = ystart
        for i, line in enumerate(svglines):
            xloc = x
            xloc += {'left': 0,
                     'right': -linewidths[i],
                     'center': -linewidths[i]/2}.get(halign, 0)

            xmin = min(xmin, xloc)
            xmax = max(xmax, xloc+linewidths[i])

            # Include extra height of tall math expressions
            drop = 0
            if i > 0:
                try:
                    drop = min(p.node.bbox.ymin for p in line if isinstance(p, Math))
                except ValueError:
                    drop = 0

            yloc -= drop
            ymin = min(ymin, yloc-lineheights[i])
            ymax = max(ymax, yloc-lineofsts[i])
            for part, size in zip(line, linesizes[i]):
                part.drawon(svgelm, xloc, yloc)
                xloc += size[0]
            yloc += lineheights[i] * self.linespacing

        if self.rotation:
            costh = cos(radians(self.rotation))
            sinth = sin(radians(self.rotation))
            p1 = (xmin-x, ymin-y)  # Corners relative to rotation point
            p2 = (xmax-x, ymin-y)
            p3 = (xmax-x, ymax-y)
            p4 = (xmin-x, ymax-y)
            x1 = x + (p1[0]*costh + p1[1]*sinth)
            x2 = x + (p2[0]*costh + p2[1]*sinth)
            x3 = x + (p3[0]*costh + p3[1]*sinth)
            x4 = x + (p4[0]*costh + p4[1]*sinth)
            y1 = y - (p1[0]*sinth - p1[1]*costh)
            y2 = y - (p2[0]*sinth - p2[1]*costh)
            y3 = y - (p3[0]*sinth - p3[1]*costh)
            y4 = y - (p4[0]*sinth - p4[1]*costh)
            bbox = (min(x1, x2, x3, x4), max(x1, x2, x3, x4),
                    min(y1, y2, y3, y4), max(y1, y2, y3, y4))

            xform = ''
            if self.rotation_mode == 'default':
                dx = {'left': x - bbox[0],
                      'right': x - bbox[1],
                      'center': x - (bbox[1]+bbox[0])/2}.get(halign, 0)
                dy = {'top': y - bbox[2],
                      'bottom': y - bbox[3],
                      'base': -sinth*dx,
                      'center': y - (bbox[3]+bbox[2])/2}.get(valign, 0)
                xform = f'translate({dx} {dy})'
                bbox = (bbox[0]+dx, bbox[1]+dx,
                        bbox[2]+dy, bbox[3]+dy)

            xform += f' rotate({-self.rotation} {x} {y})'
            if config.debug.bbox:
                rect = ET.SubElement(svg, 'rect')
                rect.attrib['x'] = fmt(bbox[0])
                rect.attrib['y'] = fmt(bbox[2])
                rect.attrib['width'] = fmt(bbox[1]-bbox[0])
                rect.attrib['height'] = fmt(bbox[3]-bbox[2])
                rect.attrib['fill'] = 'none'
                rect.attrib['stroke'] = 'red'

            svgelm.set('transform', xform)
            xmin, xmax, ymin, ymax = bbox

        return svgelm, (xmin, xmax, ymin, ymax)

    def getsize(self):
        ''' Get pixel width and height of Text. '''
        svg = ET.Element('svg')
        _, (xmin, xmax, ymin, ymax) = self._drawon(svg)
        return (xmax-xmin, ymax-ymin)

    def bbox(self):
        ''' Get bounding box (xmin, xmax, ymin, ymax) of Text. '''
        svg = ET.Element('svg')
        _, (xmin, xmax, ymin, ymax) = self._drawon(svg)
        return (xmin, xmax, ymin, ymax)


# Cache the loaded fonts to prevent reloading all the time
with pkg_resources.path('ziamath.fonts', 'STIXTwoMath-Regular.ttf') as p:
    fontname = p
loadedfonts: Dict[str, MathFont] = {'default': MathFont(fontname)}



''' Global configuration options '''
from typing import Optional
from dataclasses import dataclass, field

from ziafont import config as zfconfig


@dataclass
class DebugConfig:
    baseline: bool = False
    bbox: bool = False

    def on(self):
        self.baseline = True
        self.bbox = True

    def off(self):
        self.baseline = False
        self.bbox = False


@dataclass
class TextStyle:
    textfont: Optional[str] = None
    variant: str = 'serif'
    fontsize: float = 24
    color: str = 'black'
    linespacing: float = 1


@dataclass
class MathStyle:
    mathfont: Optional[str] = None
    variant: str = ''
    fontsize: float = 24
    color: str = 'black'
    background: str = 'none'


@dataclass
class Config:
    ''' Global configuration options for Ziamath

        Attributes
        ----------
        minsizefraction: Smallest allowed text size, as fraction of
            base size, for text such as subscripts and superscripts
        debug: Debug mode, draws bounding boxes around <mrows>
        svg2: Use SVG2.0. Disable for better browser compatibility,
            at the expense of SVG size
        precision: SVG decimal precision for coordinates
        decimal_separator: Use `.` or `,` as decimal separator. (only
            affects Latex math)
    '''
    math: MathStyle = field(default_factory=MathStyle)
    text: TextStyle = field(default_factory=TextStyle)
    minsizefraction: float = .3
    decimal_separator = '.'
    debug: DebugConfig = field(default_factory=DebugConfig)

    @property
    def svg2(self) -> bool:
        return zfconfig.svg2

    @svg2.setter
    def svg2(self, value: bool) -> None:
        zfconfig.svg2 = value

    @property
    def precision(self) -> float:
        return zfconfig.precision

    @precision.setter
    def precision(self, value: float) -> None:
        zfconfig.precision = value


config = Config()



''' <mrow> Math Elelment '''
from __future__ import annotations
from typing import Optional, Union
from copy import copy
from xml.etree import ElementTree as ET

from ziafont.fonttypes import BBox
from ziafont.glyph import SimpleGlyph, fmt

from ..drawable import Drawable
from .. import operators
from .nodetools import infer_opform, elementtext
from .mnode import Mnode


class Mrow(Mnode, tag='mrow'):
    ''' Math row, list of vertically aligned Mnodes '''
    def __init__(self, element: ET.Element, parent: 'Mnode', **kwargs):
        super().__init__(element, parent, **kwargs)
        self._setup(**kwargs)

    def _break_lines(self) -> list[list[ET.Element]]:
        ''' Break mrow into lines - to handle mspace linebreak (// in latex) '''
        lines = []
        line: list[ET.Element] = []
        for i, child in enumerate(self.element):
            if child.tag == 'mi' and elementtext(child) in operators.names:
                # Workaround for some latex2mathml operators coming back as identifiers
                child.tag = 'mo'

            if child.tag == 'mo':
                infer_opform(i, child, self)
            if child.tag == 'mspace' and child.get('linebreak', None) == 'newline':
                lines.append(line)
                line = []
            else:
                line.append(child)
        lines.append(line)
        return lines

    def _setup_multilines(self, lines: list[list[ET.Element]], **kwargs) -> None:
        ''' Multiline mrow - process each line as an mrow so we can
            get its bounding box '''
        node: Union[Mnode, Drawable]
        for i, line in enumerate(lines):
            mrowelm = ET.Element('mrow')
            mrowelm.extend(line)
            node = Mrow(mrowelm, parent=self)
            self.nodes.append(node)

        y = 0
        for i, node in enumerate(self.nodes):
            if i > 0:
                y += (node.bbox.ymax - self.nodes[i-1].bbox.ymin +
                      2 * self.units_to_points(self.font.math.consts.mathLeading))
            self.nodexy.append((0, y))
        xmin = min([n.bbox.xmin for n in self.nodes])
        xmax = max([n.bbox.xmax for n in self.nodes])
        ymin = -y+self.nodes[-1].bbox.ymin
        ymax = self.nodes[0].bbox.ymax
        self.bbox = BBox(xmin, xmax, ymin, ymax)

    def _setup_single_line(self, line: list[ET.Element], **kwargs) -> None:
        ''' Single line mrow '''
        ymax = -9999
        ymin = 9999
        height = kwargs.pop('height', None)
        i = 0
        x = xmin = xmax = 0.
        while i < len(line):
            child = line[i]
            text = elementtext(child)
            if child.tag == 'mo':
                if (text in operators.fences
                        and child.get('form') == 'prefix'
                        and child.get('stretchy') != 'false'):
                    fencekwargs = copy(kwargs)
                    j = 0
                    for j in range(i+1, len(self.element)):
                        if (self.element[j].tag == 'mo'
                                and self.element[j].get('form') == 'postfix'
                                and elementtext(self.element[j]) in operators.fences):
                            children = self.element[i+1: j]
                            fencekwargs['open'] = elementtext(child)
                            fencekwargs['close'] = elementtext(self.element[j])
                            break
                    else:  # No postfix closing fence. Enclose remainder of row.
                        children = self.element[i+1:]
                        fencekwargs['open'] = elementtext(child)
                        fencekwargs['close'] = None
                    fencekwargs['separators'] = ''
                    fenced = ET.Element('mfenced')
                    fenced.attrib.update(child.attrib)
                    fenced.attrib.update(fencekwargs)
                    fenced.extend(children)
                    node = Mnode.fromelement(fenced, parent=self, **kwargs)
                    i = j + 1

                else:
                    if text == '':
                        i += 1
                        continue  # InvisibleTimes, etc.
                    node = Mnode.fromelement(child, parent=self, height=height, **kwargs)
                    i += 1
            else:
                node = Mnode.fromelement(child, parent=self, **kwargs)
                i += 1

            self.nodes.append(node)
            self.nodexy.append((x, 0))
            xmax = max(xmax, x + node.bbox.xmax)
            xmin = min([nxy[0]+n.bbox.xmin for nxy, n in zip(self.nodexy, self.nodes)])
            ymax = max(ymax, node.bbox.ymax)
            ymin = min(ymin, node.bbox.ymin)
            x += node.xadvance()
        self.bbox = BBox(xmin, xmax, ymin, ymax)

    def _setup(self, **kwargs) -> None:
        kwargs = copy(kwargs)
        self.nodes = []

        lines = self._break_lines()
        if len(lines) > 1:
            self._setup_multilines(lines, **kwargs)
        else:
            self._setup_single_line(lines[0], **kwargs)

    def firstglyph(self) -> Optional[SimpleGlyph]:
        ''' Get the first glyph in this node '''
        i = 0
        while (i < len(self.nodes) and
               isinstance(self.nodes[i], Mnode) and
               self.nodes[i].mtag == 'mspace' and
               self.nodes[i].width <= 0):  # type: ignore
            i += 1  # Negative space shouldn't count as first glyph

        try:
            glyph = self.nodes[i].firstglyph()
        except IndexError:
            return None
        return glyph


class Merror(Mrow, tag='merror'):
    ''' Error node <merror>. Just an <mrow> with border and fill. '''
    def draw(self, x: float, y: float, svg: ET.Element) -> tuple[float, float]:
        xend, yend = super().draw(x, y, svg)
        border = ET.SubElement(svg, 'rect')
        border.set('x', fmt(x + self.bbox.xmin - 1))
        border.set('y', fmt(y - self.bbox.ymax - 1))
        border.set('width', fmt((self.bbox.xmax - self.bbox.xmin)+2))
        border.set('height', fmt((self.bbox.ymax - self.bbox.ymin)+2))
        border.set('fill', 'yellow')
        border.set('fill-opacity', '0.2')
        border.set('stroke', 'red')
        border.set('stroke-width', '1')
        return xend, yend


''' <mtable> Math Element '''
from xml.etree import ElementTree as ET
from collections import namedtuple
from copy import copy

from ziafont.fonttypes import BBox

from . import Mnode


class Mtable(Mnode, tag='mtable'):
    ''' Table node '''
    def __init__(self, element: ET.Element, parent: 'Mnode', **kwargs):
        super().__init__(element, parent, **kwargs)
        self._setup(**kwargs)

    def _setup(self, **kwargs) -> None:
        kwargs = copy(kwargs)
        rowspace = self.size_px('0.2em')
        colspace = self.size_px('0.2em')
        column_align_table = self.element.get('columnalign', 'center')

        Cell = namedtuple('Cell', 'node columnalign')

        # Build node objects from table cells
        rows = []
        for rowelm in self.element:
            assert rowelm.tag == 'mtr'
            column_align_row = rowelm.get('columnalign', column_align_table).split()

            cells = []
            for i, cellelm in enumerate(rowelm):
                assert cellelm.tag == 'mtd'

                if 'columnalign' in cellelm.attrib:
                    column_align = cellelm.get('columnalign')
                elif i < len(column_align_row):
                    column_align = column_align_row[i]
                else:  # repeat last entry of columnalign
                    column_align = column_align_row[-1]

                cells.append(Cell(Mnode.fromelement(cellelm, parent=self, **kwargs),
                                  column_align))
            rows.append(cells)

        # Compute size of each cell to size rows and columns
        rowheights = []  # Maximum height ABOVE baseline
        rowdepths = []   # Maximum distanve BELOW baseline
        for row in rows:
            rowheights.append(max([cell.node.bbox.ymax for cell in row]))
            rowdepths.append(min([cell.node.bbox.ymin for cell in row]))

        colwidths = []
        for col in [list(i) for i in zip(*rows)]:  # transposed
            colwidths.append(max([cell.node.bbox.xmax - cell.node.bbox.xmin for cell in col]))

        if self.element.get('equalrows') == 'true':
            rowheights = [max(rowheights)] * len(rows)
            rowdepths = [min(rowdepths)] * len(rows)
        if self.element.get('equalcolumns') == 'true':
            colwidths = [max(colwidths)] * len(colwidths)

        # Make Baseline of the table half the height
        # Compute baselines to each row
        totheight = sum(rowheights) - sum(rowdepths) + rowspace*(len(rows)-1)
        width = sum(colwidths) + colspace*len(colwidths)
        ytop = -totheight/2 - self.units_to_points(self.font.math.consts.axisHeight)
        baselines = []
        y = ytop
        for h, d in zip(rowheights, rowdepths):
            baselines.append(y + h)
            y += h - d + rowspace

        for r, row in enumerate(rows):
            x = colspace/2
            for c, cell in enumerate(row):
                self.nodes.append(cell.node)
                cellw = cell.node.bbox.xmax - cell.node.bbox.xmin
                if cell.columnalign == 'center':
                    xcell = x + colwidths[c]/2-cellw/2
                elif cell.columnalign == 'right':
                    xcell = x + colwidths[c]-cellw
                else:
                    xcell = x

                self.nodexy.append((xcell, baselines[r]))
                x += colwidths[c] + colspace

        ymin = min([cell.node.bbox.ymin-baselines[-1] for cell in rows[-1]])
        ymax = max([-baselines[0]+cell.node.bbox.ymax for cell in rows[0]])
        self.bbox = BBox(0, width, ymin, ymax)



''' <mspace>, <mpadded>, <mphantom> Math Elements '''
from xml.etree import ElementTree as ET

from ziafont.fonttypes import BBox

from .mnode import Mnode
from .mrow import Mrow


class Mspace(Mnode, tag='mspace'):
    ''' Blank space '''
    def __init__(self, element: ET.Element, parent: 'Mnode', **kwargs):
        super().__init__(element, parent, **kwargs)
        self.width = self.size_px(element.get('width', '0'))
        self.height = self.size_px(element.get('height', '0'))
        self.depth = self.size_px(element.get('depth', '0'))
        self._setup(**kwargs)

    def _setup(self, **kwargs) -> None:
        self.bbox = BBox(0, self.width, -self.depth, self.height)


class Mpadded(Mrow, tag='mpadded'):
    ''' Mpadded element - Mrow with extra whitespace '''
    def _setup(self, **kwargs):
        super()._setup(**kwargs)
        width = self.element.get('width', None)
        lspace = self.element.get('lspace', 0)
        height = self.element.get('height', None)
        depth = self.element.get('depth', None)
        voffset = self.element.get('voffset', 0)
        xmin, xmax, ymin, ymax = self.bbox

        def adjust(valstr: str, param: float) -> float:
            if valstr.startswith('+') or valstr.startswith('-'):
                # +X or -X means increment or decrement
                sign = valstr[0]
                if sign == '+':
                    param += self.size_px(valstr[1:])
                else:
                    param -= self.size_px(valstr[1:])
            elif valstr.endswith('%'):
                param *= float(valstr[:-1])/100
            else:
                param = self.size_px(valstr)
            return param

        if lspace:
            xshift = adjust(lspace, 0)
            for i, (x, y) in enumerate(self.nodexy):
                self.nodexy[i] = (x + xshift, y)
        if voffset:
            yshift = adjust(voffset, 0)
            for i, (x, y) in enumerate(self.nodexy):
                self.nodexy[i] = (x, y-yshift)
        if width:
            xmax = xmin + adjust(width, xmax-xmin)
        if height:
            ymax = adjust(height, ymax)
        if depth:
            ymin = -adjust(depth, ymin)
        self.bbox = BBox(xmin, xmax, ymin, ymax)


class Mphantom(Mrow, tag='mphantom'):
    ''' Phantom element. Takes up space but not drawn. '''
    def _setup(self, **kwargs) -> None:
        kwargs['phantom'] = True
        super()._setup(**kwargs)



''' <mo> Math Operator Element '''
import xml.etree.ElementTree as ET

from ziafont.fonttypes import BBox

from ..styles import styledstr
from ..drawable import Glyph
from .. import operators
from . import Mnode
from .nodetools import elementtext, subglyph


class Moperator(Mnode, tag='mo'):
    ''' Operator math element '''
    def __init__(self, element: ET.Element, parent: 'Mnode', **kwargs):
        super().__init__(element, parent, **kwargs)
        self.string = styledstr(elementtext(self.element), self.style.mathvariant)
        self.form = element.get('form', 'infix')

        # Load parameters from operators table for deciding how much space
        # to add on either side of the operator
        self.params = operators.get_params(self.string, self.form)
        self.params.update(element.attrib)
        self.width = kwargs.get('width', None)
        self.height = kwargs.get('height', None)

        minsize = self.size_px(element.get('minsize', '0'))
        maxsize = self.size_px(element.get('maxsize', '0'))
        mathsize = self.size_px(self.style.mathsize)
        if self.height:
            if minsize:
                self.height = max(self.height, minsize)
            if maxsize:
                self.height = min(self.height, maxsize)
        else:
            if mathsize:
                self.height = mathsize
            elif minsize:
                self.height = minsize
            elif maxsize:
                self.height = maxsize

        if self.string in ['|', '‖', '∣', '❘'] and self.params.get('fence') == 'true':
            # Hack around weird spacing with \middle operators
            self.params['rspace'] = 'thinmathspace'

        self._setup(**kwargs)

    def _setup(self, **kwargs):
        glyphs = [self.font.glyph(char) for char in self.string]

        if kwargs.get('sup') or kwargs.get('sub'):
            addspace = False  # Dont add lspace/rspace when in super/subscripts
            glyphs = [subglyph(g, self.font) for g in glyphs]
        else:
            addspace = True

        x = xmax = 0
        self.nodes = []

        # Add lspace
        if addspace:
            x += self.size_px(self.params.get('lspace', '0'))

        ymin = 999
        ymax = -999
        for glyph, char in zip(glyphs, self.string):
            if self.params.get('largeop') == 'true' and self.style.displaystyle:
                minh = self.font.math.consts.displayOperatorMinHeight
                glyph = self.font.math.variant(glyph.index, minh, vert=True)

            if self.width:
                glyph = self.font.math.variant(
                    glyph.index, self.points_to_units(self.width), vert=False)
            elif (self.height
                    and self.params.get('stretchy', 'false') != 'false'):
                glyph = self.font.math.variant(
                    glyph.index, self.points_to_units(self.height), vert=True)

            self.nodes.append(Glyph(
                glyph, char, self.glyphsize, self.style, **kwargs))
            self.nodexy.append((x, 0))
            xmax = max(xmax, x + self.units_to_points(glyph.path.bbox.xmax))
            x += self.units_to_points(glyph.advance())
            ymin = min(ymin, self.units_to_points(glyph.path.bbox.ymin))
            ymax = max(ymax, self.units_to_points(glyph.path.bbox.ymax))

        if addspace:
            x += self.size_px(self.params.get('rspace', '0'))
            xmax = max(xmax, x)

        try:
            xmin = self.units_to_points(glyphs[0].path.bbox.xmin)
        except IndexError:
            xmin = 0
        
        self.bbox = BBox(xmin, xmax, ymin, ymax)



''' <mroot> and <msqrt> Math Elements '''
from __future__ import annotations
from typing import Optional
import xml.etree.ElementTree as ET

from ziafont.fonttypes import BBox

from ..drawable import Glyph, HLine
from . import Mnode


class Mroot(Mnode, tag='mroot'):
    ''' Nth root '''
    def __init__(self, element: ET.Element, parent: 'Mnode', **kwargs):
        super().__init__(element, parent, **kwargs)
        self.base, self.degree = self._getbase(**kwargs)
        self._setup(**kwargs)

    def _getbase(self, **kwargs) -> tuple[Mnode, Optional[Mnode]]:
        ''' Get base and optional degree nodes for the root '''
        base = Mnode.fromelement(self.element[0], parent=self, **kwargs)
        self.increase_child_scriptlevel(self.element[1])
        degree = Mnode.fromelement(self.element[1], parent=self, **kwargs)
        return base, degree

    def _setup(self, **kwargs) -> None:
        height = self.base.bbox.ymax - self.base.bbox.ymin

        # Get the right root glyph to fit the contents
        rglyph = self.font.math.variant(self.font.glyphindex('√'),
                                        self.points_to_units(height))
        rootnode = Glyph(rglyph, '√', self.glyphsize, self.style, **kwargs)

        if self.style.displaystyle:
            verticalgap = self.units_to_points(self.font.math.consts.radicalDisplayStyleVerticalGap)
        else:
            verticalgap = self.units_to_points(self.font.math.consts.radicalVerticalGap)

        # Shift radical up/down to ensure minimum and consistent gap between top of text and overbar
        if (self.base.bbox.ymax > self.units_to_points(rglyph.bbox.ymax) - verticalgap
                or self.base.bbox.ymin <= self.units_to_points(rglyph.bbox.ymin + self.font.math.consts.radicalRuleThickness)):
            rtop = (self.base.bbox.ymax + verticalgap +
                    self.units_to_points(self.font.math.consts.radicalRuleThickness))
            yrad = -(rtop - self.units_to_points(rglyph.path.bbox.ymax))
            ytop = yrad - self.units_to_points(rglyph.path.bbox.ymax)
        else:
            yrad = 0
            ytop = -self.units_to_points(rglyph.path.bbox.ymax)

        # If the root has a degree, draw it next as it
        # determines radical x position
        self.nodes = []
        x = 0.
        if self.degree:
            x += self.units_to_points(self.font.math.consts.radicalKernBeforeDegree)
            ydeg = ytop * self.font.math.consts.radicalDegreeBottomRaisePercent/100
            self.nodes.append(self.degree)
            self.nodexy.append((x, ydeg))
            x += self.degree.xadvance()
            x += self.units_to_points(self.font.math.consts.radicalKernAfterDegree)

        self.nodes.append(rootnode)
        self.nodexy.append((x, yrad))
        x += rootnode.xadvance()
        self.nodes.append(self.base)
        self.nodexy.append((x, 0))
        width = self.base.bbox.xmax

        if (lastg := self.base.lastglyph()):
            if (italicx := self.font.math.italicsCorrection.getvalue(lastg.index)):
                width += self.units_to_points(italicx)

        self.nodes.append(HLine(
            width, self.units_to_points(self.font.math.consts.radicalRuleThickness),
            style=self.style, **kwargs))
        self.nodexy.append((x, yrad - self.units_to_points(rglyph.path.bbox.ymax)))
        xmin = self.units_to_points(rglyph.path.bbox.xmin)
        xmax = x + width
        ymin = min(-yrad + self.units_to_points(rglyph.path.bbox.ymin),
                   self.base.bbox.ymin)
        if self.degree:
            ymax = max(-yrad + self.units_to_points(rglyph.path.bbox.ymax),
                       -ydeg+self.degree.bbox.ymax)
        else:
            ymax = -yrad + self.units_to_points(rglyph.path.bbox.ymax)
        self.bbox = BBox(xmin, xmax, ymin, ymax)


class Msqrt(Mroot, tag='msqrt'):
    ''' Square root '''
    def _getbase(self, **kwargs) -> tuple[Mnode, Optional[Mnode]]:
        ''' Get base and optional degree nodes for the root '''
        if len(self.element) > 1:
            row = ET.Element('mrow')
            row.extend(list(self.element))
            base = Mnode.fromelement(row, parent=self, **kwargs)
        else:
            base = Mnode.fromelement(self.element[0], parent=self, **kwargs)
        return base, None



''' <menclose> element '''
import xml.etree.ElementTree as ET

from ziafont.fonttypes import BBox

from .. import drawable
from . import Mnode


class Menclose(Mnode, tag='menclose'):
    ''' Enclosure '''
    def __init__(self, element: ET.Element, parent: 'Mnode', **kwargs):
        super().__init__(element, parent, **kwargs)
        if len(self.element) > 1:
            row = ET.Element('mrow')
            row.extend(list(self.element))
            self.base = Mnode.fromelement(row, parent=self, **kwargs)
        else:
            self.base = Mnode.fromelement(self.element[0], parent=self, **kwargs)

        self.notation = element.get('notation', 'box').split()
        self._setup(**kwargs)

    def _setup(self, **kwargs) -> None:
        pad = 2 * self.units_to_points(self.font.math.consts.radicalRuleThickness)
        height = self.base.bbox.ymax - self.base.bbox.ymin + pad * 2
        width = self.base.bbox.xmax - self.base.bbox.xmin + pad * 2
        lw = self.units_to_points(self.font.math.consts.radicalRuleThickness)
        basex = pad
        xarrow = yarrow = 0.

        if 'box' in self.notation:
            self.nodes.append(drawable.Box(width, height, lw, style=self.style, **kwargs))
            self.nodexy.append((0, -self.base.bbox.ymax+height-pad))
        if 'circle' in self.notation:
            self.nodes.append(drawable.Ellipse(width, height, lw, style=self.style, **kwargs))
            self.nodexy.append((0, -self.base.bbox.ymax+height-pad))
        if 'roundedbox' in self.notation:
            self.nodes.append(drawable.Box(width, height, lw, style=self.style,
                                           cornerradius=lw*4, **kwargs))
            self.nodexy.append((0, -self.base.bbox.ymax+height-pad))

        if ('top' in self.notation
                or 'longdiv' in self.notation
                or 'actuarial' in self.notation):
            self.nodes.append(drawable.HLine(width, lw, style=self.style, **kwargs))
            self.nodexy.append((0, -self.base.bbox.ymax-pad))
        if ('bottom' in self.notation
                or 'madruwb' in self.notation
                or 'phasorangle' in self.notation):
            self.nodes.append(drawable.HLine(width, lw, style=self.style, **kwargs))
            self.nodexy.append((0, -self.base.bbox.ymin+pad))
        if ('right' in self.notation
                or 'madruwb' in self.notation
                or 'actuarial' in self.notation):
            self.nodes.append(drawable.VLine(height, lw, style=self.style, **kwargs))
            self.nodexy.append((self.base.bbox.xmax+pad*2, -self.base.bbox.ymax-pad))
        if ('left' in self.notation
                or 'longdiv' in self.notation):
            self.nodes.append(drawable.VLine(height, lw, style=self.style, **kwargs))
            self.nodexy.append((0, -self.base.bbox.ymax-pad))
        if 'verticalstrike' in self.notation:
            self.nodes.append(drawable.VLine(height, lw, style=self.style, **kwargs))
            self.nodexy.append((width/2, -self.base.bbox.ymax-pad))
        if 'horizontalstrike' in self.notation:
            self.nodes.append(drawable.HLine(width, lw, style=self.style, **kwargs))
            self.nodexy.append((0, -self.base.bbox.ymin-height/2))

        if 'updiagonalstrike' in self.notation:
            self.nodes.append(drawable.Diagonal(width, -height, lw, style=self.style, **kwargs))
            self.nodexy.append((0, -self.base.bbox.ymin-height+pad))
        if 'downdiagonalstrike' in self.notation:
            self.nodes.append(drawable.Diagonal(width, height, lw, style=self.style, **kwargs))
            self.nodexy.append((0, -self.base.bbox.ymin+pad))
        if 'phasorangle' in self.notation:
            self.nodes.append(drawable.Diagonal(height/3, -height, lw, style=self.style, **kwargs))
            self.nodexy.append((0, -self.base.bbox.ymin-height+pad))
            basex += height/4  # Shift base right a bit so it fits under angle

        if 'updiagonalarrow' in self.notation:
            diag = drawable.Diagonal(width, -height, lw, style=self.style, arrow=True, **kwargs)
            self.nodes.append(diag)
            self.nodexy.append((0, -self.base.bbox.ymin-height+pad))
            xarrow = diag.arroww
            yarrow = diag.arrowh

        self.nodes.append(self.base)
        self.nodexy.append((basex, 0))

        self.bbox = BBox(0, basex+width+xarrow, self.base.bbox.ymin-pad, height-pad+yarrow)



''' <mn> number math element '''
import xml.etree.ElementTree as ET

from ziafont.fonttypes import BBox

from ..styles import styledstr, auto_italic
from ..drawable import Glyph
from .nodetools import subglyph, elementtext
from .mnode import Mnode


class Mnumber(Mnode, tag='mn'):
    ''' Mnumber node <mn> '''
    def __init__(self, element: ET.Element, parent: 'Mnode', **kwargs):
        super().__init__(element, parent, **kwargs)
        self.string = self._getstring()
        self._setup(**kwargs)

    def _getstring(self) -> str:
        ''' Get the styled string for this node '''
        text = elementtext(self.element)
        return styledstr(text, self.style.mathvariant)

    def _setup(self, **kwargs) -> None:
        ymin = 9999.
        ymax = -9999.
        x = 0.

        if (leftsibling := self.leftsibling()) and leftsibling.mtag == 'mfenced':
            x = self.size_px('verythinmathspace')

        for i, char in enumerate(self.string):
            glyph = self.font.glyph(char)
            if kwargs.get('sup') or kwargs.get('sub'):
                glyph = subglyph(glyph, self.font)

            self.nodes.append(
                Glyph(glyph, char, self.glyphsize, self.style, **kwargs))

            if self.nodes[-1].bbox.xmin < 0:
                # don't let glyphs run together if xmin < 0
                x -= self.nodes[-1].bbox.xmin
            

            self.nodexy.append((x, 0))
            nextglyph = self.font.glyph(self.string[i+1]) if i < len(self.string)-1 else None
            x += self.units_to_points(glyph.advance(nextchr=nextglyph))
            ymin = min(ymin, self.units_to_points(glyph.path.bbox.ymin))
            ymax = max(ymax, self.units_to_points(glyph.path.bbox.ymax))

        try:
            xmin = self.nodes[0].bbox.xmin
            xmax = self.nodexy[-1][0] + max(self.nodes[-1].bbox.xmax,
                                            self.units_to_points(glyph.advance()))
        except IndexError:
            xmin = 0.
            xmax = x
        self.bbox = BBox(xmin, xmax, ymin, ymax)


class Midentifier(Mnumber, tag='mi'):
    ''' Number node <mn> '''
    def _getstring(self) -> str:
        ''' Get the styled string for the identifier. Applies
            italics if single-char identifier, and extra whitespace
            if function (eg 'sin')
        '''
        text = elementtext(self.element)
        if (len(text) == 1
                and not self.style.mathvariant.italic
                and not self.style.mathvariant.normal
                and auto_italic(text)):
            self.style.mathvariant.italic = True

        if len(text) > 1:
            text = '\U00002009' + text
            if self.parent.mtag not in ['msub', 'msup', 'msubsup']:
                text = text + '\U00002009'

        return styledstr(text, self.style.mathvariant)


class Mtext(Mnumber, tag='mtext'):
    ''' Text Node <mtext> '''
    def _getstring(self) -> str:
        string = ''
        if self.element.text:
            # Don't use elementtext() since it strips whitespace
            string = styledstr(self.element.text, self.style.mathvariant)
        return string

    def _setup(self, **kwargs) -> None:
        self.font.language('DFLT', '')  # Allow standard kerning to apply
        super()._setup(**kwargs)
        self.font.language('math', '')



''' Common tools for evaluating nodes '''
from typing import Union, TYPE_CHECKING
from xml.etree import ElementTree as ET

from ziafont.glyph import SimpleGlyph

from ..mathfont import MathFont

if TYPE_CHECKING:
    from .mnode import Mnode
    from ..drawable import Drawable


def infer_opform(i: int, child: ET.Element, mrow: 'Mnode') -> None:
    ''' Infer form (prefix, postfix, infix) of operator child within
        element mrow. Appends 'form' attribute to child.

        Args:
            i: Index of child within parent
            child: XML element of child
            mrow: Mnode of parent element
    '''
    # Infer form for operators
    if 'form' not in child.attrib:
        if mrow.element.tag in ['msub', 'msup', 'msubsup']:
            form = 'prefix'
        elif len(mrow.element) == 1:
            form = 'none'
        elif i == 0:
            form = 'prefix'
        elif i == len(mrow.element) - 1:
            form = 'postfix'
        else:
            form = 'infix'
        child.set('form', form)


def elementtext(element: ET.Element) -> str:
    ''' Get text of XML element '''
    try:
        txt = element.text.strip()  # type: ignore
    except AttributeError:
        txt = ''
    return txt


def node_is_singlechar(node: Union['Mnode', 'Drawable']) -> bool:
    ''' Node is a single character '''
    if node.mtag == 'mrow':
        if len(node.element) > 1:  # type:ignore
            return False
        else:
            return node_is_singlechar(node.nodes[0])  # type:ignore
    return hasattr(node, 'string') and len(node.string) == 1  # type:ignore


def subglyph(glyph: SimpleGlyph, font: MathFont) -> SimpleGlyph:
    ''' Substitute glyphs using font GSUB ssty feature. This
        substitutes glyphs like \prime for use in sub/superscripts.
    '''
    if font.gsub:
        glyphids = font.gsub.sub([glyph.index], font.features)
        if glyphids[0] != glyph.index:
            glyph = font.glyph_fromid(glyphids[0])
    return glyph



''' Read Math Table from Open Type Font file

    Reference: https://docs.microsoft.com/en-us/typography/opentype/spec/math
'''

from __future__ import annotations
from typing import Union, Sequence, Optional, TYPE_CHECKING
from collections import namedtuple
from dataclasses import dataclass
import xml.etree.ElementTree as ET

from ziafont.gpos import Coverage
from ziafont.fontread import FontReader
from ziafont.glyph import SimpleGlyph, CompoundGlyph
from ziafont.fonttypes import BBox

if TYPE_CHECKING:
    from .mathfont import MathFont


GlyphType = Union[SimpleGlyph, CompoundGlyph]

MathKernInfoRecord = namedtuple(
    'MathKernInfoRecord', ['topright', 'topleft', 'bottomright', 'bottomleft'])
GlyphPartRecord = namedtuple(
    'GlyphPartRecord', ['glyphId', 'startConnectorLength', 'endConnectorLength',
                        'fullAdvance', 'partFlags'])


def read_valuerecord(fontfile: FontReader) -> int:
    ''' Read a Math Value Record. deviceOffset is ignored. '''
    value = fontfile.readint16()
    fontfile.readint16()  # ignore
    return value


@dataclass
class MathConstants:
    ''' Data from the Math Constants Table '''
    scriptPercentScaleDown: int
    scriptScriptPercentScaleDown: int
    delimitedSubFormulaMinHeight: int
    displayOperatorMinHeight: int
    mathLeading: int
    axisHeight: int
    accentBaseHeight: int
    flattenedAccentBaseHeight: int
    subscriptShiftDown: int
    subscriptTopMax: int
    subscriptBaselineDropMin: int
    superscriptShiftUp: int
    superscriptShiftUpCramped: int
    superscriptBottomMin: int
    superscriptBaselineDropMax: int
    subSuperscriptGapMin: int
    superscriptBottomMaxWithSubscript: int
    spaceAfterScript: int
    upperLimitGapMin: int
    upperLimitBaselineRiseMin: int
    lowerLimitGapMin: int
    lowerLimitBaselineDropMin: int
    stackTopShiftUp: int
    stackTopDisplayStyleShiftUp: int
    stackBottomShiftDown: int
    stackBottomDisplayStyleShiftDown: int
    stackGapMin: int
    stackDisplayStyleGapMin: int
    stretchStackTopShiftUp: int
    stretchStackBottomShiftDown: int
    stretchStackGapAboveMin: int
    stretchStackGapBelowMin: int
    fractionNumeratorShiftUp: int
    fractionNumeratorDisplayStyleShiftUp: int
    fractionDenominatorShiftDown: int
    fractionDenominatorDisplayStyleShiftDown: int
    fractionNumeratorGapMin: int
    fractionNumDisplayStyleGapMin: int
    fractionRuleThickness: int
    fractionDenominatorGapMin: int
    fractionDenomDisplayStyleGapMin: int
    skewedFractionHorizontalGap: int
    skewedFractionVerticalGap: int
    overbarVerticalGap: int
    overbarRuleThickness: int
    overbarExtraAscender: int
    underbarVerticalGap: int
    underbarRuleThickness: int
    underbarExtraDescender: int
    radicalVerticalGap: int
    radicalDisplayStyleVerticalGap: int
    radicalRuleThickness: int
    radicalExtraAscender: int
    radicalKernBeforeDegree: int
    radicalKernAfterDegree: int
    radicalDegreeBottomRaisePercent: int


class MathTable:
    ''' OpenType Font MATH Table, provides font-specific kerning adjustments and
        glyph variants for math type setting.

        Args:
            font: Math Font
    '''
    def __init__(self, font: 'MathFont'):
        self.font = font
        self.ofst = font.tables['MATH'].offset
        self.fontfile = self.font.fontfile
        self.fontfile.seek(self.ofst)
        self.vermajor = self.fontfile.readuint16()
        self.verminor = self.fontfile.readuint16()
        assert self.vermajor == 1
        constsofst = self.fontfile.readuint16()
        glyphofst = self.fontfile.readuint16()
        variantsofst = self.fontfile.readuint16()
        self._readconsts(constsofst)
        self._readglyphinfo(glyphofst)
        self._readvariants(variantsofst)

    def _readconsts(self, ofst: int) -> None:
        ''' Read math constants table '''
        self.fontfile.seek(self.ofst + ofst)
        self.consts = MathConstants(
            scriptPercentScaleDown=self.fontfile.readint16(),
            scriptScriptPercentScaleDown=self.fontfile.readint16(),
            delimitedSubFormulaMinHeight=self.fontfile.readuint16(),
            displayOperatorMinHeight=self.fontfile.readuint16(),
            mathLeading=read_valuerecord(self.fontfile),
            axisHeight=read_valuerecord(self.fontfile),
            accentBaseHeight=read_valuerecord(self.fontfile),
            flattenedAccentBaseHeight=read_valuerecord(self.fontfile),
            subscriptShiftDown=read_valuerecord(self.fontfile),
            subscriptTopMax=read_valuerecord(self.fontfile),
            subscriptBaselineDropMin=read_valuerecord(self.fontfile),
            superscriptShiftUp=read_valuerecord(self.fontfile),
            superscriptShiftUpCramped=read_valuerecord(self.fontfile),
            superscriptBottomMin=read_valuerecord(self.fontfile),
            superscriptBaselineDropMax=read_valuerecord(self.fontfile),
            subSuperscriptGapMin=read_valuerecord(self.fontfile),
            superscriptBottomMaxWithSubscript=read_valuerecord(self.fontfile),
            spaceAfterScript=read_valuerecord(self.fontfile),
            upperLimitGapMin=read_valuerecord(self.fontfile),
            upperLimitBaselineRiseMin=read_valuerecord(self.fontfile),
            lowerLimitGapMin=read_valuerecord(self.fontfile),
            lowerLimitBaselineDropMin=read_valuerecord(self.fontfile),
            stackTopShiftUp=read_valuerecord(self.fontfile),
            stackTopDisplayStyleShiftUp=read_valuerecord(self.fontfile),
            stackBottomShiftDown=read_valuerecord(self.fontfile),
            stackBottomDisplayStyleShiftDown=read_valuerecord(self.fontfile),
            stackGapMin=read_valuerecord(self.fontfile),
            stackDisplayStyleGapMin=read_valuerecord(self.fontfile),
            stretchStackTopShiftUp=read_valuerecord(self.fontfile),
            stretchStackBottomShiftDown=read_valuerecord(self.fontfile),
            stretchStackGapAboveMin=read_valuerecord(self.fontfile),
            stretchStackGapBelowMin=read_valuerecord(self.fontfile),
            fractionNumeratorShiftUp=read_valuerecord(self.fontfile),
            fractionNumeratorDisplayStyleShiftUp=read_valuerecord(self.fontfile),
            fractionDenominatorShiftDown=read_valuerecord(self.fontfile),
            fractionDenominatorDisplayStyleShiftDown=read_valuerecord(self.fontfile),
            fractionNumeratorGapMin=read_valuerecord(self.fontfile),
            fractionNumDisplayStyleGapMin=read_valuerecord(self.fontfile),
            fractionRuleThickness=read_valuerecord(self.fontfile),
            fractionDenominatorGapMin=read_valuerecord(self.fontfile),
            fractionDenomDisplayStyleGapMin=read_valuerecord(self.fontfile),
            skewedFractionHorizontalGap=read_valuerecord(self.fontfile),
            skewedFractionVerticalGap=read_valuerecord(self.fontfile),
            overbarVerticalGap=read_valuerecord(self.fontfile),
            overbarRuleThickness=read_valuerecord(self.fontfile),
            overbarExtraAscender=read_valuerecord(self.fontfile),
            underbarVerticalGap=read_valuerecord(self.fontfile),
            underbarRuleThickness=read_valuerecord(self.fontfile),
            underbarExtraDescender=read_valuerecord(self.fontfile),
            radicalVerticalGap=read_valuerecord(self.fontfile),
            radicalDisplayStyleVerticalGap=read_valuerecord(self.fontfile),
            radicalRuleThickness=read_valuerecord(self.fontfile),
            radicalExtraAscender=read_valuerecord(self.fontfile),
            radicalKernBeforeDegree=read_valuerecord(self.fontfile),
            radicalKernAfterDegree=read_valuerecord(self.fontfile),
            radicalDegreeBottomRaisePercent=self.fontfile.readint16())

    def _readglyphinfo(self, ofst: int) -> None:
        ''' Read glyph info table '''
        self.fontfile.seek(self.ofst + ofst)
        # Read table offsets
        italics = self.fontfile.readuint16()
        topaccent = self.fontfile.readuint16()
        extendshape = self.fontfile.readuint16()
        kernofst = self.fontfile.readuint16()

        # Italics table
        self.fontfile.seek(self.ofst + ofst + italics)
        covofst = self.fontfile.readuint16()
        cnt = self.fontfile.readuint16()
        italicscorrections = []
        for i in range(cnt):
            italicscorrections.append(read_valuerecord(self.fontfile))
        italicscoverage = Coverage(self.ofst+ofst+italics+covofst,
                                   self.fontfile, nulltable=(italics==0))

        # Top Accent Attachment Table
        self.fontfile.seek(self.ofst + ofst + topaccent)
        covofst = self.fontfile.readuint16()
        cnt = self.fontfile.readuint16()
        accents = []
        for i in range(cnt):
            accents.append(read_valuerecord(self.fontfile))
        accentcoverage = Coverage(
            self.ofst+ofst+topaccent+covofst, self.fontfile, nulltable=(topaccent==0))

        # Extended Shape Coverage
        extshapes = Coverage(
            self.ofst+ofst+extendshape, self.fontfile, nulltable=(extendshape==0))

        # Kern Info
        if kernofst:
            kerninfo = MathKernInfoTable(
                self.ofst+ofst+kernofst, self.fontfile, nulltable=(kernofst==0))
        else:
            kerninfo = None

        self.italicsCorrection = MathSubTable(italicscorrections, italicscoverage)
        self.topAccentAttachment = MathSubTable(accents, accentcoverage)
        self._extendedShapeCoverage = extshapes
        self.kernInfo = kerninfo

    def _readvariants(self, ofst: int) -> None:
        ''' Read the variants table '''
        self.fontfile.seek(self.ofst + ofst)
        minoverlap = self.fontfile.readuint16()
        vertcovofst = self.fontfile.readuint16()
        horzcovofst = self.fontfile.readuint16()
        vertcount = self.fontfile.readuint16()
        horzcount = self.fontfile.readuint16()

        vertConstruction = []
        for i in range(vertcount):
            vofst = self.fontfile.readuint16()
            vertConstruction.append(
                MathConstructionTable(self.ofst+ofst+vofst, self.font, vert=True))
        horzConstruction = []
        for i in range(horzcount):
            hofst = self.fontfile.readuint16()
            horzConstruction.append(
                MathConstructionTable(self.ofst+ofst+hofst, self.font, vert=False))

        vertcoverage = Coverage(self.ofst + ofst + vertcovofst, self.fontfile)
        horzcoverage = Coverage(self.ofst + ofst + horzcovofst, self.fontfile)
        self._variantsvert = MathVariants(
            vertcoverage, vertConstruction, minoverlap, self.font, vert=True)
        self._variantshorz = MathVariants(
            horzcoverage, horzConstruction, minoverlap, self.font, vert=False)

    def kernsuper(self, glyph1: GlyphType, glyph2: GlyphType) -> tuple[int, int]:
        ''' Calculate superscript kerning between the two glyphs

            Args:
                glyph1: Last glyph in base
                glyph2: First glyph in superscript

            Returns:
                kern: Kerning shift to apply in x direction
                shift: Upward shift of superscript with respect to baseline
        '''
        if self.kernInfo is None:
            return 0, max(self.consts.superscriptShiftUp,
                          self.consts.superscriptBottomMin)

        glyph1_kern = self.kernInfo.glyph(glyph1.index)
        glyph2_kern2 = self.kernInfo.glyph(glyph2.index)

        # Extended shape, need to raise superscript up, but only if we're using
        # a variant
        if self._extendedShapeCoverage.covidx(glyph1.index) is not None:
            shiftup = glyph1.path.bbox.ymax - self.consts.superscriptShiftUp/2
        else:
            shiftup = self.consts.superscriptShiftUp

        height1 = shiftup + glyph2.path.bbox.ymin * self.consts.scriptPercentScaleDown/100
        height2 = glyph1.path.bbox.ymax - shiftup
        kern1 = kern2 = 0
        if glyph1_kern:
            kern1 += glyph1_kern.topright.getkern(height1)
            kern2 += glyph1_kern.topright.getkern(height2)
        if glyph2_kern2:
            kern1 += glyph2_kern2.bottomleft.getkern(height1)
            kern2 += glyph2_kern2.bottomleft.getkern(height2)
        return min(kern1, kern2), shiftup

    def kernsub(self, glyph1: GlyphType, glyph2: GlyphType) -> tuple[int, int]:
        ''' Calculate subscript kerning

            Args:
                glyph1: Last glyph in base
                glyph2: First glyph in subscript

            Returns:
                kern: Kerning shift to apply in x direction
                shift: Downward shift of subscript with respect to baseline
        '''
        if self.kernInfo is None:
            return 0, max(self.consts.subscriptTopMax,
                          self.consts.subscriptShiftDown)

        glyph1_kern = self.kernInfo.glyph(glyph1.index)
        glyph2_kern = self.kernInfo.glyph(glyph2.index)

        # Correction heights
        shiftdn = self.consts.subscriptShiftDown - glyph1.path.bbox.ymin
        height1 = -shiftdn + glyph2.path.bbox.ymax * self.consts.scriptPercentScaleDown/100
        height2 = glyph1.path.bbox.ymin + shiftdn
        kern1 = kern2 = 0
        if glyph1_kern:
            kern1 += glyph1_kern.bottomright.getkern(height1)
            kern2 += glyph1_kern.bottomright.getkern(height2)
        if glyph2_kern:
            kern1 += glyph2_kern.topleft.getkern(height1)
            kern2 += glyph2_kern.topleft.getkern(height2)
        return min(kern1, kern2), shiftdn

    def variant(self, glyphid: int, height: float, vert: bool = True) -> GlyphType:
        ''' Get a height variant for the glyph

            glyphid: Glyph index
            height: Desired height of glyph in font units
            vert: Vertical variant or horizontal variant
        '''
        if vert:
            variant = self._variantsvert.getvariant(glyphid, height)
        else:
            variant = self._variantshorz.getvariant(glyphid, height)
        return variant

    def variant_minmax(self, glyphid: int, ymin: float, ymax: float,
                       vert: bool = True) -> GlyphType:
        ''' Get a height variant for the glyph that covers or exceeds the range (ymin, ymax)

            Args:
                glyphid: Glyph index
                ymin: Minimum y value to contain
                ymax: Maximum y value to contain
                vert: Vertical variant or horizontal variant
        '''
        if vert:
            variant = self._variantsvert.getvariant_minmax(glyphid, ymin=ymin, ymax=ymax)
        else:
            variant = self._variantshorz.getvariant_minmax(glyphid, ymin=ymin, ymax=ymax)
        return variant

    def isextended(self, glyphid: int) -> bool:
        ''' Determine if glyph is an extended shape (has stretchy variants) '''
        return self._extendedShapeCoverage.covidx(glyphid) is not None

    def topattachment(self, glyphid: int) -> Optional[float]:
        ''' Get x-position to align an accent over the given base glyph ''' 
        return self.topAccentAttachment.getvalue(glyphid)


class MathConstructionTable:
    ''' Math Construction Table, listing size variants for a glyph

        Args:
            ofst: Byte offset to table in font file
            font: Font
            vert: Vertical or horizontal variant
    '''
    def __init__(self, ofst: int, font: 'MathFont', vert: bool = True):
        fileptr = font.fontfile.tell()
        font.fontfile.seek(ofst)
        assemblyofst = font.fontfile.readuint16()
        varcount = font.fontfile.readuint16()
        self.variants = {}
        for i in range(varcount):
            varglyph = font.fontfile.readuint16()
            advmeas = font.fontfile.readuint16()
            self.variants[advmeas] = varglyph

        self.assembly: Optional[MathAssembly]
        if assemblyofst:
            self.assembly = MathAssembly(ofst+assemblyofst, font, vert=vert)
        else:
            self.assembly = None
        font.fontfile.seek(fileptr)


class AssembledGlyph(SimpleGlyph):
    ''' Assembled glyph from Math Assembly Table

        Args:
            index: glyph index
            glyphs: list of glyphs included in the assembly
            offsets: offset (either vertical or horizontal) for each
                glyph in the assembly
            vert: Vertical or horizontal assembly
            font: Font instance
    '''
    def __init__(self,
                 index: int,
                 glyphs: Sequence[GlyphType],
                 offsets: Sequence[float],
                 vert: bool,
                 font: 'MathFont'):
        self.index = index
        self.glyphs = glyphs
        self.offsets = offsets
        self.vert = vert

        if self.vert:
            xmin = min([g.path.bbox.xmin for g in glyphs])
            xmax = max([g.path.bbox.xmax for g in glyphs])
            ymin = int(glyphs[0].path.bbox.ymin + offsets[0])
            ymax = int(glyphs[-1].path.bbox.ymax + offsets[-1])
        else:
            xmin = int(glyphs[0].path.bbox.xmin)
            xmax = int(glyphs[-1].path.bbox.xmax + offsets[-1])
            ymin = min([g.path.bbox.ymin for g in glyphs])
            ymax = max([g.path.bbox.ymax for g in glyphs])
        bbox = BBox(xmin, xmax, ymin, ymax)
        super().__init__(index, [], bbox, font)

    def advance(self, nextchr=None) -> float:
        ''' X-advance '''
        return self.bbox.xmax

    def svgpath(self, x0: float = 0, y0: float = 0,
                scale_factor: float = 1) -> Optional[ET.Element]:
        ''' Get svg <path> element for glyph, normalized to 12-point font '''
        element = ET.Element('g')

        for glyph, ofst in zip(self.glyphs, self.offsets):
            path = ''
            for operator in glyph.operators:
                if self.vert:
                    operator = operator.xform(1, 0, 0, 1, 0, ofst, 1, 1)
                else:
                    operator = operator.xform(1, 0, 0, 1, ofst, 0, 1, 1)
                segment = operator.path(x0, y0, scale=self.funits_to_points(1, scale_factor))
                if segment[0] == 'M' and path != '':
                    path += 'Z '  # Close intermediate segments
                path += segment
            if path == '':
                continue  # Don't add empty path
            path += 'Z '
            element.append(ET.Element('path', attrib={'d': path}))
        return element


class MathAssembly:
    ''' Math assembly table, for combining several glyphs to extend the size
        beyond the font built-in glyphs (for example, making a really big
        curly brace)

        Args:
            ofst: Byte offset into font file
            font: Font
            vert: Vertical or horizontal variant
    '''
    def __init__(self, ofst: int, font: 'MathFont', vert: bool = True):
        self.vert = vert
        self.font = font
        font.fontfile.seek(ofst)
        self.italicscorrection = read_valuerecord(font.fontfile)
        partcnt = font.fontfile.readuint16()
        self.parts = []
        for i in range(partcnt):
            self.parts.append(GlyphPartRecord(
                font.fontfile.readuint16(),
                font.fontfile.readuint16(),
                font.fontfile.readuint16(),
                font.fontfile.readuint16(),
                font.fontfile.readuint16()))

    def assemble(self, reqsize: float, minoverlap: float) -> AssembledGlyph:
        ''' Build glyph assembly, combining parts to create any required size

            Args:
                reqsize: Desired glyph size
                minoverlap: Minimum overlap from variants table
        '''
        size = 0.
        num_extenders = 0

        # Determine number of extender parts needed.
        while size < reqsize:
            # Min overlap give largest size with this many extenders

            testparts = []
            for part in self.parts:
                if part.partFlags:
                    testparts.extend([part]*num_extenders)
                else:
                    testparts.append(part)

            y = 0.
            for i, part in enumerate(testparts):
                if i > 0:
                    y -= minoverlap
                y += part.fullAdvance
            size = y
            num_extenders += 1

        # Decrease overlap since full extenders make it too tall
        dy = (size - reqsize) / (len(testparts)-1)

        # Build transforms for compound glyph
        offsets = []
        if self.vert:
            y = -reqsize/2 + self.font.math.consts.axisHeight
        else:
            y = 0.
        for i, part in enumerate(testparts):
            if i > 0:
                y -= minoverlap + dy
            if self.vert:
                offsets.append(y - self.font.glyph_fromid(part.glyphId).bbox.ymin)
            else:
                offsets.append(y)
            y += part.fullAdvance

        # Put together the CompoundGlyph
        glyphs = [self.font.glyph_fromid(part.glyphId) for part in testparts]

        # Make a unique ID, negative so it can't clash with other glyphs
        glyfid = -(testparts[0].glyphId + int(reqsize) << 16)
        return AssembledGlyph(glyfid, glyphs, offsets, vert=self.vert, font=self.font)


class MathVariants:
    ''' Math Variants Table. Provides different glyphs for different size symbols,
        such as curly braces and square roots.
    '''
    def __init__(self, coverage: Coverage, construction: Sequence[MathConstructionTable],
                 minoverlap: float, font: 'MathFont', vert: bool):
        self.coverage = coverage
        self.construction = construction
        self.minoverlap = minoverlap
        self.font = font
        self.vert = vert

    def getvariant(self, glyphid: int, size: float) -> GlyphType:
        ''' Get the proper size variant for the glyphid '''
        glf: GlyphType
        covidx = self.coverage.covidx(glyphid)
        if covidx is None:
            # Not covered by a variant
            return self.font.glyph_fromid(glyphid)

        construction = self.construction[covidx]
        variants = construction.variants
        try:
            variantid = variants[min(k for k in variants if k >= size)]

        except ValueError:
            # size is bigger than all variants. Use assembly table.
            if construction.assembly:
                glf = construction.assembly.assemble(size, self.minoverlap)
            else:
                glf = self.font.glyph_fromid(variants[max(variants.keys())])
        else:
            glf = self.font.glyph_fromid(variantid)

        return glf

    def getvariant_minmax(self, glyphid: int, ymin: float, ymax: float) -> GlyphType:
        ''' Get the smallest variant that encloses ymin and ymax '''
        glf: GlyphType
        covidx = self.coverage.covidx(glyphid)
        if covidx is None:
            # Not covered by a variant
            return self.font.glyph_fromid(glyphid)

        construction = self.construction[covidx]
        variants = construction.variants

        gids = list(variants.values())
        ymins = [self.font.glyph_fromid(gid).bbox.ymin for gid in gids]
        ymaxs = [self.font.glyph_fromid(gid).bbox.ymax for gid in gids]
        for gid, ymn, ymx in zip(gids, ymins, ymaxs):
            if ymn <= ymin and ymx >= ymax:
                glf = self.font.glyph_fromid(gid)
                break
        else:
            height = ymax-ymin
            if height > ymaxs[-1] - ymins[-1] and construction.assembly:
                glf = construction.assembly.assemble(height, self.minoverlap)
            else:
                glf = self.font.glyph_fromid(gids[-1])
        return glf


class MathKernTable:
    ''' Math Kerning Table, for adjusting sub/superscripts

        Args:
            ofst: Byte offset into font file
            fontfile: font file
    '''
    def __init__(self, ofst: int, fontfile: FontReader):
        fileptr = fontfile.tell()
        fontfile.seek(ofst)
        heightcnt = fontfile.readuint16()
        self.heights = []
        for i in range(heightcnt):
            self.heights.append(read_valuerecord(fontfile))
        self.kernvalues = []
        for i in range(heightcnt+1):
            self.kernvalues.append(read_valuerecord(fontfile))
        fontfile.seek(fileptr)

    def getkern(self, height: float) -> int:
        ''' Get kerning for this height '''
        if len(self.heights) == 0:
            return self.kernvalues[0]

        for i, h in enumerate(self.heights):
            if h > height:
                break
        else:
            i += 1
        return self.kernvalues[i]


class ZeroKern:
    ''' A kerning table with 0 kerning '''
    def getkern(self, height: float) -> int:
        ''' Get kerning for this height '''
        return 0


class MathKernInfoTable:
    ''' Math Kerning Info Table, listing of MathKernTables

        Args:
            ofst: Byte offset into font file
            fontfile: Font File
    '''
    def __init__(self, ofst: int, fontfile: FontReader, nulltable: bool = False):
        if nulltable:
            self.kerninfo = None
            self.coverage = None
        else:
            fontfile.seek(ofst)
            covofst = fontfile.readuint16()
            cnt = fontfile.readuint16()
            self.kerninfo = []
            for i in range(cnt):
                tr = fontfile.readuint16()
                tl = fontfile.readuint16()
                br = fontfile.readuint16()
                bl = fontfile.readuint16()
                self.kerninfo.append(MathKernInfoRecord(
                    MathKernTable(tr+ofst, fontfile) if tr else ZeroKern(),
                    MathKernTable(tl+ofst, fontfile) if tl else ZeroKern(),
                    MathKernTable(br+ofst, fontfile) if br else ZeroKern(),
                    MathKernTable(bl+ofst, fontfile) if bl else ZeroKern()))
            self.coverage = Coverage(ofst+covofst, fontfile)

    def glyph(self, glyphid: int) -> Optional[MathKernInfoRecord]:
        ''' Get kerning info record for this glyph '''
        if self.coverage is None:
            return None

        idx = self.coverage.covidx(glyphid)
        if idx is not None:
            return self.kerninfo[idx]
        return None


class MathSubTable:
    ''' Math Sub Table - generic list of values associated with glyphs '''
    def __init__(self, values: Sequence[float], coverage: Coverage):
        self.values = values
        self.coverage = coverage

    def getvalue(self, glyphid: int) -> Optional[float]:
        ''' Get a value from the table for the glyph '''
        idx = self.coverage.covidx(glyphid)
        if idx is not None:
            return self.values[idx]
        return None



''' <mfenced> math element '''
from typing import Union
from itertools import chain, zip_longest
import xml.etree.ElementTree as ET

from ziafont.fonttypes import BBox

from .. import operators
from ..drawable import Glyph
from .mnode import Mnode
from .msubsup import Msub, Msup, Msubsup
from .mfrac import Mfrac


class Mfenced(Mnode, tag='mfenced'):
    ''' Mfence element. Puts contents in parenthesis or other fence glyphs, with
        optional separators between components.
    '''
    def __init__(self, element: ET.Element, parent: 'Mnode', **kwargs):
        super().__init__(element, parent, **kwargs)
        self.openchr = element.get('open', '(')
        self.closechr = element.get('close', ')')
        self.separators = element.get('separators', ',').replace(' ', '')
        self._setup(**kwargs)

    def _setup(self, **kwargs) -> None:
        separator_elms = [ET.fromstring(f'<mo>{k}</mo>') for k in self.separators]
        fencedelms: Union[list[ET.Element], ET.Element] = []
        # Insert separators
        if len(self.element) > 1 and len(self.separators) > 0:
            fencedelms = list(chain.from_iterable(zip_longest(
                    self.element, separator_elms, fillvalue=separator_elms[-1])))   # flatten
            fencedelms = fencedelms[:len(self.element)*2-1]  # Remove any separators at end
        else:
            # Single element in fence, no separators
            fencedelms = self.element

        mrowelm = ET.Element('mrow')
        mrowelm.extend(fencedelms)

        # Make a copy of mrowelm because it can get modified
        # and we need the original later
        mrow = Mnode.fromelement(ET.fromstring(ET.tostring(mrowelm)), parent=self)
        # standard size fence glyph
        openglyph = self.font.glyph(self.openchr)
        mglyph = Glyph(
            openglyph, self.openchr, self.glyphsize, self.style, **kwargs)
        if len(mrow.nodes) == 0:
            # Opening fence with nothing in it
            fencebbox = mglyph.bbox
            xadvance = mglyph.xadvance()
        else:
            # height-adjusted fence glyph variant
            openglyph = self.font.math.variant_minmax(openglyph.index,
                                                      self.points_to_units(mrow.bbox.ymin),
                                                      self.points_to_units(mrow.bbox.ymax))
            oglyph = Glyph(openglyph, self.openchr, self.glyphsize,
                           self.style, **kwargs)

            if self.closechr:
                closeglyph = self.font.glyph(self.closechr)
                closeglyph = self.font.math.variant_minmax(closeglyph.index,
                                                           self.points_to_units(mrow.bbox.ymin),
                                                           self.points_to_units(mrow.bbox.ymax))
                cglyph = Glyph(closeglyph, self.closechr, self.glyphsize,
                               self.style, **kwargs)

            # Rebuild the mrow with the height parameter to get stretchy
            # \middle fences
            mrowelm = ET.Element('mrow')
            mrowelm.extend(fencedelms)
            mrow = Mnode.fromelement(mrowelm, parent=self, height=oglyph.bbox.ymax-oglyph.bbox.ymin)
            fencebbox = mrow.bbox
            xadvance = mrow.xadvance()

        self.nodes = []
        x = yofst = base = 0.
        yglyphmin = yglyphmax = 0.
        try:
            if self.parent.leftsibling():
                x += self.size_px('verythinmathspace')
        except AttributeError:
            pass

        if self.openchr:
            params = operators.get_params(self.openchr, 'prefix')
            rspace = self.size_px(params.get('rspace', '0'))
            self.nodes.append(oglyph)
            self.nodexy.append((x, yofst))
            x += self.units_to_points(openglyph.advance())
            x += rspace
            yglyphmin = min(-yofst+oglyph.bbox.ymin, yglyphmin)
            yglyphmax = max(-yofst+oglyph.bbox.ymax, yglyphmax)

        if len(fencedelms) > 0:
            self.nodes.append(mrow)
            self.nodexy.append((x, base))
            x += xadvance

        if self.closechr:
            try:
                # Mfrac, Msub adds space to right, remove it for fence
                if isinstance(mrow.nodes[-1], (Msub, Msup, Msubsup)):
                    x -= self.units_to_points(self.font.math.consts.spaceAfterScript)
                elif isinstance(mrow.nodes[-1], Mfrac):
                    x -= self.size_px('verythinmathspace')
            except (IndexError, AttributeError):
                pass

            if (lastg := mrow.lastglyph()):
                if (italicx := self.font.math.italicsCorrection.getvalue(lastg.index)):
                    x += mrow.units_to_points(italicx)
                        
            params = operators.get_params(self.closechr, 'postfix')
            lspace = self.size_px(params.get('lspace', '0'))
            x += lspace

            self.nodes.append(cglyph)
            self.nodexy.append((x, yofst))
            x += self.units_to_points(closeglyph.advance())
            yglyphmin = min(-yofst+cglyph.bbox.ymin, yglyphmin)
            yglyphmax = max(-yofst+cglyph.bbox.ymax, yglyphmax)

        self.bbox = BBox(0, x, min(yglyphmin, fencebbox.ymin), max(yglyphmax, fencebbox.ymax))



''' Command line interface to ziamath.

    python -m ziamath
'''

import sys
import argparse
from xml.etree.ElementTree import ParseError

import ziamath as zm


def main():
    parser = argparse.ArgumentParser(
        description='Convert MathML and Latex to standalone SVG')

    parser.add_argument(
        'file',
        nargs='?',
        help='Input file containing MathML or Latex. If empty stdin is used',
        type=argparse.FileType('r'),
        default=sys.stdin)

    parser.add_argument(
        '-o',
        dest='output',
        nargs='?',
        help='Output file. If empty, stdout is used',
        type=argparse.FileType('w'),
        default=sys.stdout)

    parser.add_argument(
        '--latex',
        help='Use Latex input mode',
        action='store_true')

    parser.add_argument(
        '--svg1',
        help='Keep SVG1.x compatibility',
        action='store_true')

    parser.add_argument(
        '--font',
        '-f',
        default=None,
        help='Font file (TTF/OTF). Must contain MATH table')

    parser.add_argument(
        '--size',
        '-s',
        type=int,
        default=None,
        help='Font size in points')

    parser.add_argument(
        '--precision',
        '-p',
        type=int,
        default=None,
        help='Decimal precision for SVG coordinates')

    parser.add_argument(
        '--version',
        action='version',
        version=zm.__version__)

    args = parser.parse_args()
    if args.file.isatty():
        parser.print_help()
        return 0

    if args.svg1:
        zm.config.svg2 = False
    if args.precision:
        zm.config.precision = args.precision

    kwargs = {'font': args.font,
              'size': args.size}

    mathinput = args.file.read()    
    if args.latex:
        svg = zm.Latex(mathinput, **kwargs).svg()
    else:
        try:
            svg = zm.Math(mathinput, **kwargs).svg()
        except ParseError:
            svg = zm.Latex(mathinput, **kwargs).svg()

    print(svg, file=args.output)
    return 0


if __name__ == '__main__':
    sys.exit(main())



''' Math node - parent class of all math nodes '''
from __future__ import annotations
from typing import Optional, MutableMapping, Type
import logging
from xml.etree import ElementTree as ET

from ziafont.fonttypes import BBox
from ziafont.glyph import SimpleGlyph, fmt

from ..mathfont import MathFont
from ..drawable import Drawable
from ..styles import MathStyle, parse_style
from ..config import config
from .. import operators
from .nodetools import elementtext, infer_opform

_node_classes: dict[str, Type['Mnode']] = {}


class Mnode(Drawable):
    ''' Math Drawing Node

        Args:
            element: XML element for the node
            size: base font size in points
            parent: Mnode of parent
    '''
    mtag = 'mnode'

    def __init__(self, element: ET.Element, parent: 'Mnode', **kwargs):
        super().__init__()
        self.element = element
        self.font: MathFont = parent.font
        self.parent = parent
        self.size: float = parent.size
        self.style: MathStyle = parse_style(self.element, parent.style)
        self.params: MutableMapping[str, str] = {}
        self.nodes: list[Drawable] = []
        self.nodexy: list[tuple[float, float]] = []
        self.glyphsize = max(
            self.size * (self.font.math.consts.scriptPercentScaleDown/100)**self.style.scriptlevel,
            self.font.basesize*config.minsizefraction)
        if self.style.mathsize:
            self.glyphsize = self.size_px(self.style.mathsize)

        self._font_pts_per_unit = self.size / self.font.info.layout.unitsperem
        self._glyph_pts_per_unit = self.glyphsize / self.font.info.layout.unitsperem
        self.bbox = BBox(0, 0, 0, 0)

    def __init_subclass__(cls, tag: str) -> None:
        ''' Register this subclass so fromelement() can find it '''
        _node_classes[tag] = cls
        cls.mtag = tag

    @classmethod
    def fromelement(cls, element: ET.Element, parent: 'Mnode', **kwargs) -> 'Mnode':
        ''' Construct a new node from the element and its parent '''
        if element.tag in ['math', 'mtd', 'mtr']:
            element.tag = 'mrow'
        elif element.tag == 'ms':
            element.tag = 'mtext'
        elif element.tag == 'mi' and elementtext(element) in operators.names:
            # Workaround for some latex2mathml operators coming back as identifiers
            element.tag = 'mo'

        if element.tag == 'mo':
            infer_opform(0, element, parent)

        node = _node_classes.get(element.tag, None)
        if node:
            return node(element, parent, **kwargs)

        logging.warning('Undefined element %s', element)
        return _node_classes['mrow'](element, parent, **kwargs)

    def _setup(self, **kwargs) -> None:
        ''' Calculate node position assuming this node is at 0, 0. Also set bbox. '''
        self.bbox = BBox(0, 0, 0, 0)

    def units_to_points(self, value: float) -> float:
        ''' Convert value in font units to points at this glyph size '''
        return value * self._glyph_pts_per_unit

    def font_units_to_points(self, value: float) -> float:
        ''' Convert value in font units to points at the base font size '''
        return value * self._font_pts_per_unit

    def points_to_units(self, value: float) -> float:
        ''' Convert points back to font units '''
        return value / self._glyph_pts_per_unit

    def increase_child_scriptlevel(self, element: ET.Element) -> None:
        ''' Increase the child element's script level one higher
            than this element, if not overridden in child's attributes
        '''
        element.attrib.setdefault('scriptlevel', str(self.style.scriptlevel+1))

    def leftsibling(self) -> Optional[Drawable]:
        ''' Left node sibling. The one that was just placed. '''
        try:
            node = self.parent.nodes[-1]
            if node.mtag == 'mrow' and node.nodes:
                node = node.nodes[-1]
        except (IndexError, AttributeError):
            node = None

        return node

    def firstglyph(self) -> Optional[SimpleGlyph]:
        ''' Get the first glyph in this node '''
        try:
            return self.nodes[0].firstglyph()
        except IndexError:
            return None

    def lastglyph(self) -> Optional[SimpleGlyph]:
        ''' Get the last glyph in this node '''
        try:
            return self.nodes[-1].lastglyph()
        except IndexError:
            return None

    def lastchar(self) -> Optional[str]:
        ''' Get the last character in this node '''
        try:
            return self.nodes[-1].lastchar()
        except IndexError:
            return None

    def size_px(self, size: str, fontsize: float = None) -> float:
        ''' Get size in points from the attribute string '''
        if fontsize is None:
            fontsize = self.glyphsize

        numsize = {"veryverythinmathspace": f'{1/18}em',
                   "verythinmathspace": f'{2/18}em',
                   "thinmathspace": f'{3/18}em',
                   "mediummathspace": f'{4/18}em',
                   "thickmathspace": f'{5/18}em',
                   "verythickmathspace": f'{6/18}em',
                   "veryverythickmathspace": f'{7/18}em',
                   "negativeveryverythinmathspace": f'{-1/18}em',
                   "negativeverythinmathspace": f'{-2/18}em',
                   "negativethinmathspace": f'{-3/18}em',
                   "negativemediummathspace": f'{-4/18}em',
                   "negativethickmathspace": f'{-5/18}em',
                   "negativeverythickmathspace": f'{-6/18}em',
                   "negativeveryverythickmathspace": f'{-7/18}em',
                   }.get(size, size)

        numsize = numsize.rstrip('px')
        try:
            pxsize = float(numsize)
        except ValueError as exc:
            if numsize.endswith('em'):
                pxsize = float(numsize[:-2]) * fontsize
            elif numsize.endswith('pt'):
                pxsize = float(numsize[:-2]) * 1.333  # 1.333 points to pixels
            else:
                pxsize = 0
        return pxsize

    def draw(self, x: float, y: float, svg: ET.Element) -> tuple[float, float]:
        ''' Draw the node on the SVG

            Args:
                x: Horizontal position in SVG coordinates
                y: Vertical position in SVG coordinates
                svg: SVG drawing as XML
        '''
        if config.debug.bbox:
            rect = ET.SubElement(svg, 'rect')
            rect.set('x', fmt(x + self.bbox.xmin))
            rect.set('y', fmt(y - self.bbox.ymax))
            rect.set('width', fmt((self.bbox.xmax - self.bbox.xmin)))
            rect.set('height', fmt((self.bbox.ymax - self.bbox.ymin)))
            rect.set('fill', 'none')
            rect.set('stroke', 'blue')
            rect.set('stroke-width', '0.2')
        if config.debug.baseline:
            base = ET.SubElement(svg, 'path')
            base.set('d', f'M {x} 0 L {x+self.bbox.xmax} 0')
            base.set('stroke', 'red')

        if self.style.mathbackground not in ['none', None]:
            rect = ET.SubElement(svg, 'rect')
            rect.set('x', fmt(x + self.bbox.xmin))
            rect.set('y', fmt(y - self.bbox.ymax))
            rect.set('width', fmt((self.bbox.xmax - self.bbox.xmin)))
            rect.set('height', fmt((self.bbox.ymax - self.bbox.ymin)))
            rect.set('fill', str(self.style.mathbackground))

        nodex = nodey = 0.
        for (nodex, nodey), node in zip(self.nodexy, self.nodes):
            node.draw(x+nodex, y+nodey, svg)
        return x+nodex, y+nodey



from __future__ import annotations
from typing import Optional
import math
import xml.etree.ElementTree as ET

from ziafont.fonttypes import BBox
from ziafont.glyph import SimpleGlyph, fmt
from .config import config
from .styles import MathStyle


class Drawable:
    ''' Base class for drawable nodes '''
    mtag = 'drawable'
    nodes: list[Drawable] = []

    def __init__(self):
        self.bbox = BBox(0, 0, 0, 0)

    def firstglyph(self) -> Optional[SimpleGlyph]:
        ''' Get the first glyph in this node '''
        return None

    def lastglyph(self) -> Optional[SimpleGlyph]:
        ''' Get the last glyph in this node '''
        return None

    def lastchar(self) -> Optional[str]:
        ''' Get the last character in this node '''
        return None

    def xadvance(self) -> float:
        ''' X-advance for the glyph. Usually bbox.xmax '''
        return self.bbox.xmax

    def draw(self, x: float, y: float, svg: ET.Element) -> tuple[float, float]:
        ''' Draw the element. Must be subclassed. '''
        raise NotImplementedError


class Glyph(Drawable):
    ''' A single glyph

        Args:
            glyph: The glyph to draw
            char: unicode character represented by the glyph
            size: point size
            style: font MathStyle
    '''
    def __init__(self, glyph: SimpleGlyph, char: str, size: float,
                 style: MathStyle = None, **kwargs):
        super().__init__()
        self.glyph = glyph
        self.char = char
        self.size = size
        self.phantom = kwargs.get('phantom', False)
        self.style = style if style else MathStyle()
        self._funits_to_pts = self.size / self.glyph.font.info.layout.unitsperem
        self.bbox = BBox(
            self.funit_to_points(self.glyph.path.bbox.xmin),
            self.funit_to_points(self.glyph.path.bbox.xmax),
            self.funit_to_points(self.glyph.path.bbox.ymin),
            self.funit_to_points(self.glyph.path.bbox.ymax))
        
    def funit_to_points(self, value: float) -> float:
        ''' Convert font units to SVG points '''
        return value * self._funits_to_pts

    def firstglyph(self) -> Optional[SimpleGlyph]:
        ''' Get the first glyph in this node '''
        return self.glyph

    def lastglyph(self) -> Optional[SimpleGlyph]:
        ''' Get the last glyph in this node '''
        return self.glyph

    def lastchar(self) -> Optional[str]:
        ''' Get the last character in this node '''
        return self.char

    def xadvance(self) -> float:
        ''' X-advance for the glyph. Usually bbox.xmax '''
        return self.funit_to_points(self.glyph.advance())

    def draw(self, x: float, y: float, svg: ET.Element) -> tuple[float, float]:
        ''' Draw the node on the SVG

            Args:
                x: Horizontal position in SVG coordinates
                y: Vertical position in SVG coordinates
                svg: SVG drawing as XML
        '''
        symbols = svg.findall('symbol')
        symids = [sym.attrib.get('id') for sym in symbols]
        if self.glyph.id not in symids and config.svg2:
            svg.append(self.glyph.svgsymbol())
        if not self.phantom:
            path = self.glyph.place(x, y, self.size)
            if path is not None:
                svg.append(path)
            if self.style.mathcolor and len(svg) > 0:
                svg[-1].set('fill', str(self.style.mathcolor))
        x += self.funit_to_points(self.glyph.advance())
        return x, y


class HLine(Drawable):
    ''' Horizontal Line. '''
    def __init__(self, length: float, lw: float,
                 style: MathStyle = None, **kwargs):
        super().__init__()
        self.length = length
        self.lw = lw
        self.phantom = kwargs.get('phantom', False)
        self.bbox = BBox(0, self.length, -self.lw/2, self.lw/2)
        self.style = style if style else MathStyle()

    def draw(self, x: float, y: float, svg: ET.Element) -> tuple[float, float]:
        ''' Draw the node on the SVG

            Args:
                x: Horizontal position in SVG coordinates
                y: Vertical position in SVG coordinates
                svg: SVG drawing as XML
        '''
        if not self.phantom:
            # Use rectangle so it can change color with 'fill' attribute
            # and not mess up glyphs with 'stroke' attribute
            bar = ET.SubElement(svg, 'rect')
            bar.attrib['x'] = fmt(x)
            bar.attrib['y'] = fmt(y)
            bar.attrib['width'] = fmt(self.length)
            bar.attrib['height'] = fmt(self.lw)
            if self.style.mathcolor:
                bar.attrib['fill'] = str(self.style.mathcolor)
        return x+self.length, y


class VLine(Drawable):
    ''' Vertical Line. '''
    def __init__(self, height: float, lw: float,
                 style: MathStyle = None, **kwargs):
        super().__init__()
        self.height = height
        self.lw = lw
        self.phantom = kwargs.get('phantom', False)
        self.bbox = BBox(0, self.lw, 0, self.height)
        self.style = style if style else MathStyle()

    def draw(self, x: float, y: float, svg: ET.Element) -> tuple[float, float]:
        ''' Draw the node on the SVG

            Args:
                x: Horizontal position in SVG coordinates
                y: Vertical position in SVG coordinates
                svg: SVG drawing as XML
        '''
        if not self.phantom:
            # Use rectangle so it can change color with 'fill' attribute
            # and not mess up glyphs with 'stroke' attribute
            bar = ET.SubElement(svg, 'rect')
            bar.attrib['x'] = fmt(x-self.lw/2)
            bar.attrib['y'] = fmt(y)
            bar.attrib['width'] = fmt(self.lw)
            bar.attrib['height'] = fmt(self.height)
            if self.style.mathcolor:
                bar.attrib['fill'] = str(self.style.mathcolor)
        return x, y


class Box(Drawable):
    ''' Box '''
    def __init__(self, width: float, height: float, lw: float,
                 cornerradius: float = None,
                 style: MathStyle = None, **kwargs):
        super().__init__()
        self.width = width
        self.height = height
        self.cornerradius = cornerradius
        self.lw = lw
        self.phantom = kwargs.get('phantom', False)
        self.bbox = BBox(0, self.width, 0, self.height)
        self.style = style if style else MathStyle()

    def draw(self, x: float, y: float, svg: ET.Element) -> tuple[float, float]:
        ''' Draw the node on the SVG

            Args:
                x: Horizontal position in SVG coordinates
                y: Vertical position in SVG coordinates
                svg: SVG drawing as XML
        '''
        if not self.phantom:
            bar = ET.SubElement(svg, 'rect')
            bar.set('x', fmt(x))
            bar.set('y', fmt(y-self.height))
            bar.set('width', fmt(self.width))
            bar.set('height', fmt(self.height))
            bar.set('stroke-width', fmt(self.lw))
            bar.set('stroke', self.style.mathcolor)
            bar.set('fill', self.style.mathbackground)
            if self.cornerradius:
                bar.set('rx', fmt(self.cornerradius))

        return x+self.width, y


class Diagonal(Drawable):
    ''' Diagonal Line - corners of Box '''
    def __init__(self, width: float, height: float, lw: float,
                 arrow: bool = False,
                 style: MathStyle = None, **kwargs):
        super().__init__()
        self.width = width
        self.height = height
        self.lw = lw
        self.arrow = arrow
        self.phantom = kwargs.get('phantom', False)
        self.bbox = BBox(0, self.width, 0, self.height)
        self.style = style if style else MathStyle()

        self.arroww = self.width
        self.arrowh = self.height
        if self.arrow:
            # Bbox needs to be a bit bigger to accomodate arrowhead
            theta = math.atan2(-self.height, self.width)
            self.arroww = (10+self.lw*2) * math.cos(theta)
            self.arrowh = (10+self.lw*2) * math.sin(theta)

    def draw(self, x: float, y: float, svg: ET.Element) -> tuple[float, float]:
        ''' Draw the node on the SVG

            Args:
                x: Horizontal position in SVG coordinates
                y: Vertical position in SVG coordinates
                svg: SVG drawing as XML
        '''
        if not self.phantom:
            bar = ET.SubElement(svg, 'path')
            if self.arrow:
                arrowdef = ET.SubElement(svg, 'defs')
                marker = ET.SubElement(arrowdef, 'marker')
                marker.set('id', 'arrowhead')
                marker.set('markerWidth', '10')
                marker.set('markerHeight', '7')
                marker.set('refX', '0')
                marker.set('refY', '3.5')
                marker.set('orient', 'auto')
                poly = ET.SubElement(marker, 'polygon')
                poly.set('points', '0 0 10 3.5 0 7')

            bar.set('d', f'M {fmt(x)} {fmt(y-self.height)} L {fmt(x+self.width)} {fmt(y)}')
            bar.set('stroke-width', fmt(self.lw))
            bar.set('stroke', self.style.mathcolor)
            if self.arrow:
                bar.set('marker-end', 'url(#arrowhead)')

        return x+self.width, y


class Ellipse(Drawable):
    ''' Ellipse '''
    def __init__(self, width: float, height: float, lw: float,
                 style: MathStyle = None, **kwargs):
        super().__init__()
        self.width = width
        self.height = height
        self.lw = lw
        self.phantom = kwargs.get('phantom', False)
        self.bbox = BBox(0, self.width, 0, self.height)
        self.style = style if style else MathStyle()

    def draw(self, x: float, y: float, svg: ET.Element) -> tuple[float, float]:
        ''' Draw the node on the SVG

            Args:
                x: Horizontal position in SVG coordinates
                y: Vertical position in SVG coordinates
                svg: SVG drawing as XML
        '''
        if not self.phantom:
            bar = ET.SubElement(svg, 'ellipse')
            bar.set('cx', fmt(x+self.width/2))
            bar.set('cy', fmt(y-self.height/2))
            bar.set('rx', fmt(self.width/2))
            bar.set('ry', fmt(self.height/2))
            bar.set('stroke-width', fmt(self.lw))
            bar.set('stroke', self.style.mathcolor)
            bar.set('fill', self.style.mathbackground)
        return x+self.width, y
